# Copyright 2022-Present CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import jax.numpy as jnp

def time_values(fs, T, initial_time=0, endpoint=False):
    """Returns a sequence of time values sampled at a specific frequency for a specific duration 

    Args:
        fs (float): Sample rate of signal in Hz.
        T (float): Period of the signal in seconds.
        initial_time (float): time at waveform start in seconds, default is 0.
        endpoint (bool): Whether to include last end point in the sequence or not, default is False.

    Returns:
        jax.numpy.ndarray: A 1D array of time values

    Example:
        ::

            >>> fs=2 # Hz
            >>> T = 4 # Sec
            >>> cr.nimble.dsp.time_values(fs, T)
            DeviceArray([0. , 0.5, 1. , 1.5, 2. , 2.5, 3. , 3.5], dtype=float32)
            >>> cr.nimble.dsp.time_values(fs, T, endpoint=True)
            DeviceArray([0. , 0.5, 1. , 1.5, 2. , 2.5, 3. , 3.5, 4. ], dtype=float32)
            >>> cr.nimble.dsp.time_values(fs, T, initial_time=-2, endpoint=True)
            DeviceArray([-2. , -1.5, -1. , -0.5,  0. ,  0.5,  1. ,  1.5,  2. ], dtype=float32)
    """
    # Number of samples
    n = int(fs * T) + int(endpoint)
    # Points in time where the chirp will be computed.
    t = jnp.linspace(initial_time, initial_time+T, n, endpoint=endpoint)
    return t

def chirp(fs, T, f0, f1, initial_phase=0):
    """Generates a frequency sweep from low to high over time.

    Args:
        fs (float): Sample rate of signal in Hz.
        T (float): Period of the signal in seconds.
        f0 (float): Start (lower) frequency of chirp in Hz.
        f1 (float): Stop (upper) frequency of chirp in Hz.
        initial_phase (float): phase at t=0 in radians, default is 0.

    Returns:
        A tuple comprising (i) an array of time values in seconds and (ii) an array of signal values

    Adapted from https://udel.edu/~mm/gr/chirp.py
    """
    # Chirp rate in Hz/s.
    c = (f1 - f0) / T 
    # Number of samples
    n = int(fs * T)
    # Points in time where the chirp will be computed.
    t = jnp.linspace(0, T, n, endpoint=False)
    # Instantaneous phase in Hz is integral of frequency, f(t) = ct + f0.
    phase_hz = (c * t**2) / 2 + (f0 * t)
    # Convert to radians.
    phase_rad = 2 * jnp.pi * phase_hz 
    # Offset by user-specified initial phase
    phase_rad += initial_phase
    # compute the chirp signal at the specified points in time
    signal = jnp.cos(phase_rad)
    return t, signal


def chirp_centered(fs, T, fc, bw, initial_phase=0):
    """Generates a frequency sweep from low to high over time defined by central frequency and bandwidth.

    Args:
        fs (float): Sample rate of signal in Hz.
        T (float): Period of the signal in seconds.
        fc (float): Central frequency of chirp in Hz.
        bw (float): Bandwidth (end frequency -  start frequency) of chirp in Hz.
        initial_phase (float): phase at t=0 in radians, default is 0.

    Returns:
        A tuple comprising (i) an array of time values in seconds and (ii) an array of signal values

    Adapted from https://udel.edu/~mm/gr/chirp.py
    """
    f0 = fc - bw / 2.
    f1 = fc + bw / 2.
    return chirp(fs, T, f0, f1, initial_phase)



def pulse(fs, T, start_time, end_time, initial_time=0):
    """Generates a pulse signal which is 1 between start and end times and 0 everwhere else

    Args:
        fs (float): Sample rate of signal in Hz.
        T (float): Period of the signal in seconds.
        start_time (float): Start time of the box signal in seconds
        end_time (float): End time of the box signal in seconds
        initial_time (float): time at waveform start in seconds, default is 0.

    Returns:
        A tuple comprising (i) an array of time values in seconds and (ii) an array of signal values
    """
    t = time_values(fs, T, initial_time)
    signal = jnp.zeros_like(t)
    index = jnp.logical_and(t >= start_time, t < end_time)
    signal = signal.at[index].set(1)
    return t, signal

def gaussian(fs, T, b, a=1., initial_time=0):
    """Generates a Gaussian signal

    Args:
        fs (float): Sample rate of signal in Hz.
        T (float): Period of the signal in seconds.
        b (float): The location (in time) where the pulse is centered in seconds.
        a (float): scale of Gaussian signal (in seconds).
        initial_time (float): time at waveform start in seconds, default is 0.

    Returns:
        A tuple comprising (i) an array of time values in seconds and (ii) an array of signal values

    """
    t = time_values(fs, T, initial_time)
    a = jnp.atleast_2d(jnp.asarray(a)).T
    tb = t - b
    # square the scale s^2
    wsq = a**2
    # t^2
    xsq = tb**2
    # the gaussian term e^{-t^2/2a^2}
    gauss = jnp.exp(-xsq / (2 * wsq))
    return t, jnp.squeeze(gauss)

def decaying_sine_wave(fs, T, f, alpha, initial_phase=0, initial_time=0):
    """Generates a decaying sinusoid
    Args:
        fs (float): Sample rate of signal in Hz.
        T (float): Period of the signal in seconds.
        f (float): Frequency of the sine wave in Hz.
        alpha (float): Exponential decay factor in Hz.
        initial_phase (float): phase at t=0 in radians, default is 0.
        initial_time (float): time at waveform start in seconds, default is 0.

    Returns:
        A tuple comprising (i) an array of time values in seconds and (ii) an array of signal values
    """
    t = time_values(fs, T, initial_time)
    phase = 2*jnp.pi*f*t
    phase += initial_phase
    decay = jnp.exp(-alpha*t)
    signal = decay * jnp.sin(phase) 
    return t, signal

def transient_sine_wave(fs, T, f, start_time, end_time, initial_phase=0, initial_time=0):
    """Generates a transient sinusoid between start and end times

    Args:
        fs (float): Sample rate of signal in Hz.
        T (float): Period of the signal in seconds.
        f (float): Frequency of the sine wave in Hz.
        start_time (float): Start time of the sine wave in seconds
        end_time (float): End time of the sine wave in seconds
        initial_phase (float): phase at t=0 in radians, default is 0.
        initial_time (float): time at waveform start in seconds, default is 0.

    Returns:
        A tuple comprising (i) an array of time values in seconds and (ii) an array of signal values

    Example:
        ::

            fs = 100
            T = 16
            f = 2
            start_time = 2
            end_time = 6
            initial_time = -4
            t, signal = transient_sine_wave(fs, T, f, start_time, end_time, initial_time=initial_time)        
    """
    t = time_values(fs, T, initial_time)
    phase = 2*jnp.pi*f*t
    phase += initial_phase
    signal = jnp.sin(phase)
    mask = jnp.logical_or(t < start_time, t >= end_time)
    signal = signal.at[mask].set(0)
    return t, signal

def gaussian_pulse(fs, T, b, fc=1000, bw=0.5, bwr=-6, retquad=False, retenv=False, initial_time=0):
    """Generates a Gaussian modulated sinusoid

    Args:
        fs (float): Sample rate of signal in Hz.
        T (float): Period of the signal in seconds.
        b (float): The location (in time) where the pulse is centered in seconds.
        fc (float): Center frequency of the Gaussian Pulse in Hz.
        bw (float): Fractional bandwidth in frequency domain of pulse.
        bwr (float): Reference level at which fractional bandwidth is calculated (dB).
        retquad (bool): Include/exclude quadrature(imaginary) part of the signal in the result
        retenv (bool): Include/exclude the Gaussian envelope of the signal in the result
        initial_time (float): time at waveform start in seconds, default is 0.

    Returns:
        tuple: A tuple comprising of 
        (i) t: time values in seconds at which the signal is computed.
        (ii) real_part: values of the real part of the signal at these time values.
        (iii) imag_part: values of the quadrature/imaginary part of the signal at these time values (if retquad is True).
        (iv) envelope: values of the Gaussian envelope of the signal at these time values (if retenv is True).

    Adapted from https://github.com/scipy/scipy/blob/v1.7.1/scipy/signal/waveforms.py#L161-L258
    """
    t = time_values(fs, T, initial_time)
    ref = math.pow(10.0, bwr / 20.0)
    a = -(jnp.pi * fc * bw) ** 2 / (4.0 * math.log(ref))
    tb = t - b
    envelope = jnp.exp(-a * tb * tb)
    real_part = envelope * jnp.cos(2 * jnp.pi * fc * tb)
    if retquad:
        imag_part = envelope * jnp.sin(2 * jnp.pi * fc * tb)
    if retquad:
        if retenv:
            return t, real_part, imag_part, envelope
        else:
            return t, real_part, imag_part
    else:
        if retenv:
            return t, real_part, envelope
        else:
            return t, real_part




# Copyright 2022-Present CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Discrete Cosine Transforms

Adapted from:

* http://www-personal.umich.edu/~mejn/computational-physics/dcst.py
* https://dsp.stackexchange.com/questions/2807/fast-cosine-transform-via-fft
"""

from jax import jit
import jax.numpy as jnp
import jax.numpy.fft as jfft

def dct(y):
    """Computes the 1D Type-II DCT transform

    Args:
        y (jax.numpy.ndarray): The 1D real signal

    Returns:
        jax.numpy.ndarray: The Type-II Discrete Cosine Transform coefficients of y
    """
    n = y.shape[0]
    y2 = jnp.concatenate( (y[:],  y[::-1]))
    c = jfft.rfft(y2, axis=0)[:n]
    ks = jnp.arange(n)
    phi = jnp.exp(-1j*jnp.pi*ks/(2*n))
    prod = (phi*c.T).T
    return jnp.real(prod)


def idct(a):
    """Computes the 1D Type-II Inverse DCT transform

    Args:
        a (jax.numpy.ndarray): The Type-II DCT transform coefficients of a 1D real signal

    Returns:
        jax.numpy.ndarray: The 1D real signal y s.t. a = dct(y)
     """
    n = a.shape[0]
    shape = (1,)+a.shape[1:]
    ks = jnp.arange(n)
    phi = jnp.exp(1j*jnp.pi*ks/(2*n))
    upper = (phi*a.T).T
    lower = jnp.zeros(shape)
    c = jnp.concatenate((upper, lower))
    return jfft.irfft(c, axis=0)[:n]


def orthonormal_dct(y):
    """Computes the 1D Type-II DCT transform such that the transform is orthonormal

    Args:
        y (jax.numpy.ndarray): The 1D real signal

    Returns:
        jax.numpy.ndarray: The orthonormal Type-II Discrete Cosine Transform coefficients of y 

    Orthonormality ensures that

    .. math::

        \\langle a, a \\rangle = \\langle y, y \\rangle
    """
    n = y.shape[0]
    factor = jnp.sqrt(1/(2*n))
    ks = jnp.arange(n)
    phi = jnp.exp(-1j*jnp.pi*ks/(2*n))
    # scaling to make the transform orthonormal
    phi = phi.at[0].set(phi[0]*1/jnp.sqrt(2))
    phi = phi * factor

    y2 = jnp.concatenate( (y[:],  y[::-1]))
    c = jfft.rfft(y2, axis=0)[:n]
    prod = jnp.real(phi*c.T).T
    # phi = phi*jnp.sqrt(2)/n 
    return prod

def orthonormal_idct(a):
    """Computes the 1D Type-II IDCT transform such that the transform is orthonormal

    Args:
        a (jax.numpy.ndarray): The orthonormal Type-II DCT transform coefficients of a 1D real signal

    Returns:
        jax.numpy.ndarray: The 1D real signal y s.t. a = orthonormal_dct(y)
    """
    n = a.shape[0]
    factor = jnp.sqrt(2*n)
    ks = jnp.arange(n)

    phi = jnp.exp(1j*jnp.pi*ks/(2*n))
    # scaling to make the transform orthonormal
    phi = phi*factor
    phi = phi.at[0].set(phi[0]*jnp.sqrt(2))

    upper = (phi*a.T).T
    lower = jnp.zeros((1,)+a.shape[1:])
    c = jnp.concatenate((upper, lower))
    return jfft.irfft(c, axis=0)[:n]



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax.numpy as jnp
from jax import jit, lax

from jax.scipy.linalg import svd

"""Utilities based on the Singular Value Decomposition of a matrix
"""

def orth(A, rcond=None):
    """
    Constructs an orthonormal basis for the range of A using SVD

    Args:
        A (jax.numpy.ndarray): Input matrix of size (M, N) where 
            M is the dimension of the ambient vector space and N 
            is the number of vectors in A
        rcond (float) : Relative condition number. 
            Singular values ``s`` smaller than
            ``rcond * max(s)`` are considered zero.
            Default: floating point eps * max(M,N).

    Returns:
        (jax.numpy.ndarray, int): Returns a tuple consisting of
         * the left singular vectors of A
         * the effective rank of A 

    To get the ONB, follow the two step process::

        Q, r = orth(A)
        Q = Q[:, :r]

    Examples:
        >>> A = jnp.array([[2, 0, 0], [0, 5, 0]])  # rank 2 array
        >>> Q, rank = orth(A)
        >>> print(Q)
        [[0. 1.]
        [1. 0.]]
        >>> print(rank)
        2

    The implementation is adapted from ``scipy.linalg.orth``. 
    However, the return type is different. We return the rank
    of the matrix separately. This is done so that ``orth``
    can be JIT compiled. Dynamic slices are not supported by
    JIT.
    """
    u, s, vh = svd(A, full_matrices=False)
    rank = effective_rank_from_svd(u, s, vh)
    return u, rank

orth_jit  = jit(orth, static_argnums=(1,))


def row_space(A, rcond=None):
    """
    Constructs an orthonormal basis for the row space of A using SVD

    Args:
        A (jax.numpy.ndarray): Input matrix of size (M, N) where 
            M is the dimension of the ambient vector space and N 
            is the number of vectors in A
        rcond (float) : Relative condition number. 
            Singular values ``s`` smaller than
            ``rcond * max(s)`` are considered zero.
            Default: floating point eps * max(M,N).

    Returns:
        (jax.numpy.ndarray, int): Returns a tuple consisting of
         * the right singular vectors of A
         * the effective rank of A 

    To get the ONB for the row space, follow the two step process::

        Q, r = orth(A)
        Q = Q[:, :r]

    Examples:
        >>> A = jnp.array([[2, 0, 0], [0, 5, 0]]).T
        >>> print(A)
        [[2 0]
        [0 5]
        [0 0]]
        >>> Q, rank = crla.row_space(A)
        >>> print(Q[:, :rank])
        [[0. 1.]
        [1. 0.]]
    """
    u, s, vh = svd(A, full_matrices=False)
    rank = effective_rank_from_svd(u, s, vh)
    Q = jnp.conjugate(vh.T)
    return Q, rank

row_space_jit  = jit(row_space, static_argnums=(1,))


def null_space(A, rcond=None):
    """
    Constructs an orthonormal basis for the null space of A using SVD

    Args:
        A (jax.numpy.ndarray): Input matrix of size (M, N) where 
            M is the dimension of the ambient vector space and N 
            is the number of vectors in A
        rcond (float) : Relative condition number. 
            Singular values ``s`` smaller than
            ``rcond * max(s)`` are considered zero.
            Default: floating point eps * max(M,N).

    Returns:
        (jax.numpy.ndarray, int): Returns a tuple consisting of
         * the right singular vectors of A
         * the effective rank of A 

    To get the ONB for the null space of A, follow the two step process::

        Z, r = null_space(A)
        Z = Z[:, r:]

    The dimension of the effective null space is :math:`N - r` where r is the rank of A.

    Examples:
        >>> A = random.normal(key, (3, 5))
        >>> Z, r = null_space(A)
        >>> Z = Z[:, r:]
        >>> Z.shape
        (5, 2)
        >>> print(jnp.allclose(A @ Z, 0))
        True
    """
    u, s, vh = svd(A, full_matrices=True)
    rank = effective_rank_from_svd(u, s, vh)
    N = jnp.conjugate(vh.T)
    return N, rank

null_space_jit  = jit(null_space, static_argnums=(1,))




def left_null_space(A, rcond=None):
    """
    Constructs an orthonormal basis for the left null space of A using SVD

    Args:
        A (jax.numpy.ndarray): Input matrix of size (M, N) where 
            M is the dimension of the ambient vector space and N 
            is the number of vectors in A
        rcond (float) : Relative condition number. 
            Singular values ``s`` smaller than
            ``rcond * max(s)`` are considered zero.
            Default: floating point eps * max(M,N).

    Returns:
        (jax.numpy.ndarray, int): Returns a tuple consisting of
         * the left singular vectors of A
         * the effective rank of A 

    To get the ONB for the left null space of A, follow the two step process::

        Z, r = left_null_space(A)
        Z = Z[:, r:]

    The dimension of the effective null space is :math:`M - r` where r is the rank of A.

    Examples:
        >>> A = random.normal(key, (6, 4))
        >>> Z, r = left_null_space(A)
        >>> Z = Z[:, r:]
        >>> Z.shape
        (6, 2)
        >>> print(jnp.allclose(Z.T @ A, 0))
        True
    """
    u, s, vh = svd(A, full_matrices=True)
    rank = effective_rank_from_svd(u, s, vh)
    return u, rank

left_null_space_jit  = jit(left_null_space, static_argnums=(1,))


def effective_rank(A, rcond=None):
    """
    Returns the effective rank of A based on its singular value decomposition

    Args:
        A (jax.numpy.ndarray): Input matrix of size (M, N) where 
            M is the dimension of the ambient vector space and N 
            is the number of vectors in A
        rcond (float) : Relative condition number. 
            Singular values ``s`` smaller than
            ``rcond * max(s)`` are considered zero.
            Default: floating point eps * max(M,N).

    Returns:
        (int): Returns the effective rank of A 

    Examples:
        >>> A = random.normal(key, (3, 5))
        >>> r = svd_effective_rank(A)
        >>> print(r)
        3
    """
    u, s, vh = svd(A, full_matrices=False)
    return effective_rank_from_svd(u, s, vh, rcond)

effective_rank_jit  = jit(effective_rank, static_argnums=(1,))


def effective_rank_from_svd(u, s, vh, rcond=None):
    """Returns the effective rank of a matrix from its SVD

    Args:
        u (jax.numpy.ndarray): Left singular vectors 
        s (jax.numpy.ndarray): Singular values 
        vh (jax.numpy.ndarray): Right singular vectors (Hermitian transposed)
        rcond (float) : Relative condition number. 
            Singular values ``s`` smaller than
            ``rcond * max(s)`` are considered zero.
            Default: floating point eps * max(M,N).

    Returns:
        (int): Returns the effective rank by analyzing the singular values 

    It is assumed that the SVD has already been computed.

    Examples:
        >>> A  = random.normal(key, (6, 4))
        >>> u, s, vh = jax.scipy.linalg.svd(A)
        >>> r = crla.effective_rank_from_svd(u, s, vh)
        >>> print(r)
        4
    """
    M, N = u.shape[0], vh.shape[1]
    if rcond is None:
        rcond = jnp.finfo(s.dtype).eps * max(M, N)
    tol = jnp.amax(s) * rcond
    rank = jnp.sum(s > tol, dtype=int)
    return rank


@jit
def singular_values(A):
    """Returns the singular values of a matrix

    Args:
        A (jax.numpy.ndarray): Input matrix of size (M, N) where 
            M is the dimension of the ambient vector space and N 
            is the number of vectors in A


    Returns:
        (jax.numpy.ndarray): The list of singular values


    Examples:
        >>> key = random.PRNGKey(0)
        >>> A = random.normal(key, (20, 10))
        >>> print(singular_values(A))
        [6.6780386  6.19980196 5.65133988 4.89395458 4.49728071 3.9139061
         3.50887351 2.66701591 2.12520081 1.63708146]
    """
    return jnp.linalg.svd(A, full_matrices=False, compute_uv=False)



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp

from cr.nimble import solve_Lx_b, solve_LTx_b, solve_Ux_b, solve_UTx_b


def cholesky_update_on_add_column(A, L, I, i):
    """Incrementally updates the Cholesky factorization :math:`G = L L^T` where :math:`G = A_I^T A_I with new column A_i`

    Note:

        Assumes that the first iteration is already done.
    """
    a = A[:, i]
    v = A[:, I].T @ a
    m, n = L.shape
    z = jnp.zeros((m, 1))
    w = solve_Lx_b(L, v)
    s = jnp.sqrt(a.T @ a - w.T @ w)
    L0 = jnp.hstack((L, z))
    L1 = jnp.hstack((w.T, s))
    L = jnp.vstack((L0, L1))
    return L

def cholesky_build_factor(A):
    """Builds the Cholesky factor :math:`L` of Gram matrix :math:`G = A^T A`  as :math:`G = L L^T` incrementally column wise 
    """
    a = A[:, 0]
    L = jnp.reshape(jnp.sqrt(a @ a), (1,1))
    for j in range(1, A.shape[1]):
        I = jnp.arange(0, j)
        L = cholesky_update_on_add_column(A, L, I, j)
    return L


# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pairwise distances between a set of points
"""
from jax import jit
import jax.numpy as jnp

@jit
def pairwise_sqr_l2_distances_rw(A, B):
    r"""Computes the pairwise squared distances between points in A and points in B where each point is a row vector

    Args:
        A (jax.numpy.ndarray): A set of M K-dimensional points (row-wise)
        B (jax.numpy.ndarray): A set of N K-dimensional points (row-wise)

    Returns:
        (jax.numpy.ndarray): An MxN matrix D of squared distances 
        between points in A and points in B

    * Let the ambient space of points be :math:`\mathbb{F}^K`.
    * :math:`A` contains the points :math:`a_i` with :math:`1 \leq i \leq M` 
      and each point maps to a row of :math:`A`.
    * :math:`B` contains the points :math:`b_j` with :math:`1 \leq j \leq N` 
      and each point maps to a row of :math:`B`.

    Then the distance matrix :math:`D` is of size :math:`M \times N` and consists of:

    .. math::

        d_{i, j} = \| a_i - b_j \|_2^2 = \langle a_i - b_j , a_i - b_j  \rangle
    """
    M = A.shape[0]
    N = B.shape[0]
    # compute squared sums for each row vector
    a_sums = jnp.sum(A*A, axis=1)
    # reshape to Mx1 column vector
    a_sums = jnp.reshape(a_sums, (M, 1))
    # broadcast to MxN matrix
    a_sums = a_sums * jnp.ones((1,N))

    # compute squared sums for each row vector
    b_sums = jnp.sum(B*B, axis=1)
    # broadcast to MxN matrix
    b_sums = b_sums * jnp.ones((M, 1))

    # multiply A (M x p) and B.T (p x N)
    prods = A @ B.T 
    return a_sums + b_sums - 2 * prods

@jit
def pairwise_sqr_l2_distances_cw(A, B):
    r"""Computes the pairwise squared distances between points in A and points in B where each point is a column vector

    Args:
        A (jax.numpy.ndarray): A set of M K-dimensional points (column-wise)
        B (jax.numpy.ndarray): A set of N K-dimensional points (column-wise)

    Returns:
        (jax.numpy.ndarray): An MxN matrix D of squared distances 
        between points in A and points in B

    * Let the ambient space of points be :math:`\mathbb{F}^K`.
    * :math:`A` contains the points :math:`a_i` with :math:`1 \leq i \leq M` 
      and each point maps to a column of :math:`A`.
    * :math:`B` contains the points :math:`b_j` with :math:`1 \leq j \leq N` 
      and each point maps to a column of :math:`B`.

    Then the distance matrix :math:`D` is of size :math:`M \times N` and consists of:

    .. math::

        d_{i, j} = \| a_i - b_j \|_2^2 = \langle a_i - b_j , a_i - b_j  \rangle

    """
    M = A.shape[1]
    N = B.shape[1]
    # compute squared sums for each column vector
    a_sums = jnp.sum(A*A, axis=0)
    # reshape to Mx1 column vector
    a_sums = jnp.reshape(a_sums, (M, 1))
    # broadcast to MxN matrix
    a_sums = a_sums * jnp.ones((1,N))

    # compute squared sums for each column vector
    b_sums = jnp.sum(B*B, axis=0)
    # broadcast to MxN matrix
    b_sums = b_sums * jnp.ones((M, 1))

    # multiply A.T (M x p) and B (p x N)
    prods = A.T @ B 
    return a_sums + b_sums - 2 * prods


@jit
def pairwise_l2_distances_rw(A, B):
    r"""Computes the pairwise distances between points in A and points in B where each point is a row vector

    Args:
        A (jax.numpy.ndarray): A set of M K-dimensional points (row-wise)
        B (jax.numpy.ndarray): A set of N K-dimensional points (row-wise)

    Returns:
        (jax.numpy.ndarray): An MxN matrix D of euclidean distances 
        between points in A and points in B
    """
    return jnp.sqrt(pairwise_sqr_l2_distances_rw(A, B))


@jit
def pairwise_l2_distances_cw(A, B):
    r"""Computes the pairwise distances between points in A and points in B where each point is a column vector

    Args:
        A (jax.numpy.ndarray): A set of M K-dimensional points (column-wise)
        B (jax.numpy.ndarray): A set of N K-dimensional points (column-wise)

    Returns:
        (jax.numpy.ndarray): An MxN matrix D of euclidean distances 
        between points in A and points in B
    """
    return jnp.sqrt(pairwise_sqr_l2_distances_cw(A, B))

@jit
def pdist_sqr_l2_rw(A):
    r"""Computes the pairwise squared distances between points in A where each point is a row vector

    Args:
        A (jax.numpy.ndarray): A set of N K-dimensional points (row-wise)

    Returns:
        (jax.numpy.ndarray): An NxN matrix D of squared euclidean distances 
        between points in A

    * Let the ambient space of points be :math:`\mathbb{F}^K`.
    * :math:`A` contains the points :math:`a_i` with :math:`1 \leq i \leq N` 
      and each point maps to a row of :math:`A`.

    Then the distance matrix :math:`D` is of size :math:`N \times N` and consists of:

    .. math::

        d_{i, j} = \| a_i - a_j \|_2^2 = \langle a_i - a_j , a_i - a_j  \rangle
    """
    M = A.shape[0]
    # compute squared sums for each row vector
    sums = jnp.sum(A*A, axis=1)
    # broadcast to MxM matrix
    a_sums = jnp.reshape(sums, (M,1)) * jnp.ones((1, M))
    b_sums = sums * jnp.ones((M, 1))

    # multiply A (M x p) and A.T (p x M)
    prods = A @ A.T 
    return a_sums + b_sums - 2*prods

@jit
def pdist_sqr_l2_cw(A):
    r"""Computes the pairwise squared distances between points in A where each point is a column vector

    Args:
        A (jax.numpy.ndarray): A set of N K-dimensional points (column-wise)

    Returns:
        (jax.numpy.ndarray): An NxN matrix D of squared euclidean distances 
        between points in A

    * Let the ambient space of points be :math:`\mathbb{F}^K`.
    * :math:`A` contains the points :math:`a_i` with :math:`1 \leq i \leq N` 
      and each point maps to a column of :math:`A`.

    Then the distance matrix :math:`D` is of size :math:`N \times N` and consists of:

    .. math::

        d_{i, j} = \| a_i - a_j \|_2^2 = \langle a_i - a_j , a_i - a_j  \rangle
    """
    M = A.shape[1]
    # compute squared sums for each col vector
    sums = jnp.sum(A*A, axis=0)
    # broadcast to MxN matrix
    a_sums = jnp.reshape(sums, (M, 1)) * jnp.ones((1,M))
    b_sums = sums * jnp.ones((M, 1))
    # multiply A.T (M x p) and A (p x M)
    prods = A.T @ A 
    return a_sums + b_sums - 2 * prods

@jit
def pdist_l2_rw(A):
    r"""Computes the pairwise distances between points in A where ach point is a row vector

    Args:
        A (jax.numpy.ndarray): A set of N K-dimensional points (row-wise)

    Returns:
        (jax.numpy.ndarray): An NxN matrix D of euclidean distances 
        between points in A
    """
    return jnp.sqrt(pdist_sqr_l2_rw(A))

@jit
def pdist_l2_cw(A):
    r"""Computes the pairwise distances between points in A where each point is a column vector

    Args:
        A (jax.numpy.ndarray): A set of N K-dimensional points (column-wise)

    Returns:
        (jax.numpy.ndarray): An NxN matrix D of euclidean distances 
        between points in A
    """
    return jnp.sqrt(pdist_sqr_l2_cw(A))


@jit
def pairwise_l1_distances_rw(A, B):
    r"""Computes the pairwise city-block distances between points in A and points in B where each point is a row vector

    Args:
        A (jax.numpy.ndarray): A set of M K-dimensional points (row-wise)
        B (jax.numpy.ndarray): A set of N K-dimensional points (row-wise)

    Returns:
        (jax.numpy.ndarray): An MxN matrix D of city-block distances 
        between points in A and points in B
    """
    return jnp.sum(jnp.abs(A[:, None, :] - B[None, :, :]), axis=-1)

@jit
def pairwise_l1_distances_cw(A, B):
    r"""Computes the pairwise city-block distances between points in A and points in B where each point is a column vector

    Args:
        A (jax.numpy.ndarray): A set of M K-dimensional points (column-wise)
        B (jax.numpy.ndarray): A set of N K-dimensional points (column-wise)

    Returns:
        (jax.numpy.ndarray): An MxN matrix D of city-block distances 
        between points in A and points in B
    """
    return jnp.sum(jnp.abs(A[:, :, None] - B[:, None, :]), axis=0)

@jit
def pdist_l1_rw(A):
    r"""Computes the pairwise city-block distances between points in A where each point is a row vector

    Args:
        A (jax.numpy.ndarray): A set of M K-dimensional points (row-wise)

    Returns:
        (jax.numpy.ndarray): An MxM matrix D of city-block distances 
        between points in A
    """
    return pairwise_l1_distances_rw(A, A)

@jit
def pdist_l1_cw(A):
    r"""Computes the pairwise city-block distances between points in A where each point is a column vector

    Args:
        A (jax.numpy.ndarray): A set of M K-dimensional points (column-wise)

    Returns:
        (jax.numpy.ndarray): An MxM matrix D of city-block distances 
        between points in A
    """
    return pairwise_l1_distances_cw(A, A)


@jit
def pairwise_linf_distances_rw(A, B):
    r"""Computes the pairwise Chebyshev distances between points in A and points in B where each point is a row vector

    Args:
        A (jax.numpy.ndarray): A set of M K-dimensional points (row-wise)
        B (jax.numpy.ndarray): A set of N K-dimensional points (row-wise)

    Returns:
        (jax.numpy.ndarray): An MxN matrix D of Chebyshev distances 
        between points in A and points in B
    """
    return jnp.max(jnp.abs(A[:, None, :] - B[None, :, :]), axis=-1)

@jit
def pairwise_linf_distances_cw(A, B):
    r"""Computes the pairwise Chebyshev distances between points in A and points in B where each point is a column vector

    Args:
        A (jax.numpy.ndarray): A set of M K-dimensional points (column-wise)
        B (jax.numpy.ndarray): A set of N K-dimensional points (column-wise)

    Returns:
        (jax.numpy.ndarray): An MxN matrix D of Chebyshev distances 
        between points in A and points in B
    """
    return jnp.max(jnp.abs(A[:, :, None] - B[:, None, :]), axis=0)


@jit
def pdist_linf_rw(A):
    r"""Computes the pairwise Chebyshev distances between points in A where each point is a row vector

    Args:
        A (jax.numpy.ndarray): A set of M K-dimensional points (row-wise)

    Returns:
        (jax.numpy.ndarray): An MxM matrix D of Chebyshev distances 
        between points in A
    """
    return pairwise_linf_distances_rw(A, A)

@jit
def pdist_linf_cw(A):
    r"""Computes the pairwise Chebyshev distances between points in A where each point is a column vector

    Args:
        A (jax.numpy.ndarray): A set of M K-dimensional points (column-wise)

    Returns:
        (jax.numpy.ndarray): An MxM matrix D of Chebyshev distances 
        between points in A
    """
    return pairwise_linf_distances_cw(A, A)




# Copyright 2021 CR.Sparse Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
from jax import random, jit
import jax.numpy.fft as jfft
from jax.scipy import signal

from cr.nimble import sqr_norms_l2_cw, sqr_norms_l2_rw
from cr.nimble import is_matrix
from .discrete.number import next_pow_of_2

def find_first_signal_with_energy_le_rw(X, energy):
    """Returns the index of the first row which has energy less than the specified threshold
    """
    assert is_matrix(X)
    energies = sqr_norms_l2_rw(X)
    index = jnp.argmax(energies <= energy)
    return index if energies[index] <= energy else jnp.array(-1)

def find_first_signal_with_energy_le_cw(X, energy):
    """Returns the index of the first column which has energy less than the specified threshold
    """
    assert is_matrix(X)
    energies = sqr_norms_l2_cw(X)
    index = jnp.argmax(energies <= energy)
    return index if energies[index] <= energy else jnp.array(-1)


def randomize_rows(key, X):
    """Randomizes the rows in X

    Args:
        key: a PRNG key used as the random key.
        X (jax.numpy.ndarray): A 2D data matrix

    Returns:
        (jax.numpy.ndarray): The data matrix with randomized rows
    """
    assert is_matrix(X)
    m, n = X.shape
    r = random.permutation(key, m)
    return X[r, :]

def randomize_cols(key, X):
    """Randomizes the columns in X

    Args:
        key: a PRNG key used as the random key.
        X (jax.numpy.ndarray): A 2D data matrix

    Returns:
        (jax.numpy.ndarray): The data matrix with randomized columns
    """
    assert is_matrix(X)
    m, n = X.shape
    r = random.permutation(key, n)
    return X[:, r]


def largest_indices(x, K):
    """Returns the indices of K largest entries in x by magnitude

    Args:
        x (jax.numpy.ndarray): An data vector/point
        K (int): The number of largest entries to be identified in x

    Returns:
        (jax.numpy.ndarray): An index vector of size K identifying the K largest entries in x
        in descending order
    """
    indices = jnp.argsort(jnp.abs(x))
    return indices[:-K-1:-1]

def largest_indices_rw(X, K):
    """Returns the indices of K largest entries by magnitude in each row of X

    Args:
        X (jax.numpy.ndarray): An (S,N) data matrix with data points in rows
        K (int): The number of largest entries to be identified in each row of X

    Returns:
        (jax.numpy.ndarray): An (S,K) index matrix indices of K largest elements in each row of X
    """
    indices = jnp.argsort(jnp.abs(X), axis=1)
    return indices[:, :-K-1:-1]

def largest_indices_cw(X, K):
    """Returns the indices of K largest entries by magnitude in each column of X

    Args:
        X (jax.numpy.ndarray): An (N,S) data matrix with data points in columns
        K (int): The number of largest entries to be identified in each column of X

    Returns:
        (jax.numpy.ndarray): An (K,S) index matrix indices of K largest elements in each column of X
    """
    indices = jnp.argsort(jnp.abs(X), axis=0)
    return indices[:-K-1:-1, :]

def take_along_rows(X, indices):
    """Picks K entries from each row of X specified by indices matrix

    Args:
        X (jax.numpy.ndarray): An (S,N) data matrix with data points in rows
        indices (jax.numpy.ndarray): An (S,K) index matrix identifying the values to be picked up from X

    Returns:
        (jax.numpy.ndarray): An (S,K) data matrix subset of X containing K elements from each row of X
    """
    return jnp.take_along_axis(X, indices, axis=1)

def take_along_cols(X, indices):
    """Picks K entries from each column of X specified by indices matrix

    Args:
        X (jax.numpy.ndarray): An (N,S) data matrix with data points in columns
        indices (jax.numpy.ndarray): An (K,S) index matrix identifying the values to be picked up from X

    Returns:
        (jax.numpy.ndarray): An (K,S) data matrix subset of X containing K elements from each column of X
    """
    return jnp.take_along_axis(X, indices, axis=0)

def sparse_approximation(x, K):
    """Keeps only largest K non-zero entries by magnitude in a vector x

    Args:
        x (jax.numpy.ndarray): An data vector/point
        K (int): The number of largest entries to be kept in x

    Returns:
        (jax.numpy.ndarray): x modified so that all entries except the K largest entries are set to 0
    """
    if K == 0:
        return x.at[:].set(0)
    indices = jnp.argsort(jnp.abs(x))
    #print(x, K, indices)
    return x.at[indices[:-K]].set(0)
    
def sparse_approximation_cw(X, K):
    #return jax.vmap(sparse_approximation, in_axes=(1, None), out_axes=1)(X, K)
    """Keeps only largest K non-zero entries by magnitude in each column of X

    Args:
        X (jax.numpy.ndarray): An (N,S) data matrix with data points in columns
        K (int): The number of largest entries to be kept in each column of X

    Returns:
        (jax.numpy.ndarray): X modified so that all entries except the K largest entries are set to 0 in each column
    """
    if K == 0:
        return X.at[:].set(0)
    indices = jnp.argsort(jnp.abs(X), axis=0)
    for c in range(X.shape[1]):
        ind = indices[:-K, c]
        X = X.at[ind, c].set(0)
    return X

def sparse_approximation_rw(X, K):
    """Keeps only largest K non-zero entries by magnitude in each row of X

    Args:
        X (jax.numpy.ndarray): An (S,N) data matrix with data points in rows
        K (int): The number of largest entries to be kept in each row of X

    Returns:
        (jax.numpy.ndarray): X modified so that all entries except the K largest entries are set to 0 in each row
    """
    if K == 0:
        return X.at[:].set(0)
    indices = jnp.argsort(jnp.abs(X), axis=1)
    for r in range(X.shape[0]):
        ind = indices[r, :-K]
        X = X.at[r, ind].set(0)
    return X


def build_signal_from_indices_and_values(length, indices, values):
    """Builds a sparse signal from its non-zero entries (specified by their indices and values)

    Args:
        length (int): Length of the sparse signal
        indices (jax.numpy.ndarray): An index vector of length K identifying non-zero entries
        values (jax.numpy.ndarray): Values to be stored in the non-zero positions

    Returns:
        (jax.numpy.ndarray): Resulting sparse signal such that x[indices] == values 
    """
    x = jnp.zeros(length)
    indices = jnp.asarray(indices)
    values = jnp.asarray(values)
    return x.at[indices].set(values)


def nonzero_values(x):
    """Returns the values of non-zero entries in x

    Args:
        x (jax.numpy.ndarray): A sparse signal

    Returns:
        (jax.numpy.ndarray): The signal stripped of its zero values


    Note:
        This function cannot be JIT compiled as the size of output is data dependent.
    """
    return x[x != 0]

def nonzero_indices(x):
    """Returns the indices of non-zero entries in x

    Args:
        x (jax.numpy.ndarray): A sparse signal

    Returns:
        (jax.numpy.ndarray): The indices of nonzero entries in x

    Note:
        This function cannot be JIT compiled as the size of output is data dependent.

    See Also:
        :func:`support`
    """
    return jnp.nonzero(x)[0]


def support(x):
    """Returns the indices of non-zero entries in x

    Args:
        x (jax.numpy.ndarray): A sparse signal

    Returns:
        (jax.numpy.ndarray): The support of x a.k.a. the indices of nonzero entries in x
        
    Note:
        This function cannot be JIT compiled as the size of output is data dependent.

    See Also:
        :func:`nonzero_indices`
    """
    return jnp.nonzero(x)[0]


def hard_threshold(x, K):
    """Returns the indices and corresponding values of largest K non-zero entries in a vector x

    Args:
        x (jax.numpy.ndarray): A sparse/compressible signal
        K (int): The number of largest entries to be kept in x

    Returns:
        (jax.numpy.ndarray, jax.numpy.ndarray): A tuple comprising of:
            * The indices of K largest entries in x
            * Corresponding entries in x

    See Also:
        :func:`hard_threshold_sorted`
        :func:`hard_threshold_by`
    """
    indices = jnp.argsort(jnp.abs(x))
    I = indices[:-K-1:-1]
    x_I = x[I]
    return I, x_I

def hard_threshold_sorted(x, K):
    """Returns the sorted indices and corresponding values of largest K non-zero entries in a vector x

    Args:
        x (jax.numpy.ndarray): A sparse/compressible signal
        K (int): The number of largest entries to be kept in x

    Returns:
        (jax.numpy.ndarray, jax.numpy.ndarray): A tuple comprising of:
            * The indices of K largest entries in x sorted in ascending order
            * Corresponding entries in x

    See Also:
        :func:`hard_threshold`
    """
    # Sort entries in x by their magnitude
    indices = jnp.argsort(jnp.abs(x))
    # Pick the indices of K-largest (magnitude) entries in x (from behind)
    I = indices[:-K-1:-1]
    # Make sure that indices are sorted in ascending order
    I = jnp.sort(I)
    # Pick corresponding values
    x_I = x[I]
    return I, x_I

def hard_threshold_by(x, t):
    """
    Sets all entries in x to be zero which are less than t in magnitude

    Args:
        x (jax.numpy.ndarray): A sparse/compressible signal
        t (float): The threshold value

    Returns:
        (jax.numpy.ndarray): x modified such that all values below t are set to 0

    Note:
        This function doesn't change the length of x and can be JIT compiled

    See Also:
        :func:`hard_threshold`
    """
    valid = jnp.abs(x) >= t
    return x * valid

def largest_indices_by(x, t):
    """
    Returns the locations of all entries in x which are larger than t in magnitude

    Args:
        x (jax.numpy.ndarray): A sparse/compressible signal
        t (float): The threshold value

    Returns:
        (jax.numpy.ndarray): An index vector of all entries in x which are above the threshold

    Note:
        This function cannot be JIT compiled as the length of output is data dependent

    See Also:
        :func:`hard_threshold_by`
    """
    return jnp.where(jnp.abs(x) >= t)[0]

def dynamic_range(x):
    """Returns the ratio of largest and smallest values (by magnitude) in x (dB)

    Args:
        x (jax.numpy.ndarray): A signal

    Returns:
        (float): The dynamic range between largest and smallest value

    Note:
        This function is not suitable for sparse signals where some values are actually 0

    See Also:
        :func:`nonzero_dynamic_range`
    """
    x = jnp.sort(jnp.abs(x))
    return 20 * jnp.log10(x[-1] / x[0])


def nonzero_dynamic_range(x):
    """Returns the ratio of largest and smallest non-zero values (by magnitude) in x (dB)

    Args:
        x (jax.numpy.ndarray): A sparse/compressible signal

    Returns:
        (float): The dynamic range between largest and smallest nonzero value

    Note:
        This function cannot be JIT compiled as it depends on arrays whose size is data-dependent.

    See Also:
        :func:`dynamic_range`
    """
    x = nonzero_values(x)
    return dynamic_range(x)


def normalize(data, axis=-1):
    """Normalizes a data vector (data - mu) / sigma 

    Args:
        data (jax.numpy.ndarray): A data vector or array
        axis (int): For nd arrays, the axis along which the data normalization will be done

    Returns:
        (jax.numpy.ndarray): Normalized data vector/array
    """
    mu = jnp.mean(data, axis)
    data = data - mu
    variance = jnp.var(data, axis)
    data = data / jnp.sqrt(variance)
    return data

normalize_jit = jit(normalize, static_argnums=(1,))


def frequency_spectrum(x, dt=1.):
    """Frequency spectrum of 1D data using FFT
    """
    n = len(x)
    nn = next_pow_of_2(n)
    X = jfft.fft(x, nn)
    f = jfft.fftfreq(nn, d=dt)
    X = jfft.fftshift(X)
    f = jfft.fftshift(f)
    return f, X

def power_spectrum(x, dt=1.):
    """Power spectrum of 1D data using FFT
    """
    n = len(x)
    T = dt * n
    f, X = frequency_spectrum(x, dt)
    nn = len(f)
    n2 = nn // 2
    f = f[n2:]
    X = X[n2:]
    sxx = (X * jnp.conj(X)) / T
    sxx = jnp.abs(sxx)
    return f, sxx

def energy(data, axis=-1):
    """
    Computes the energy of the signal along the specified axis
    """
    power = jnp.abs(data) ** 2
    return jnp.sum(power, axis)


def interpft(x, N):
    """Interpolates x to n points in Fourier Transform domain
    """
    n = len(x)
    assert n < N
    a = jfft.fft(x)
    nyqst = (n + 1) // 2
    z = jnp.zeros(N -n)
    a1 = a[:nyqst+1]
    a2 = a[nyqst+1:]
    b = jnp.concatenate((a1, z, a2))
    if n % 2 == 0:
        b = b.at[nyqst].set(b[nyqst] /2 )
        b = b.at[nyqst + N -n].set(b[nyqst])
    y = jfft.ifft(b)
    if jnp.isrealobj(x):
        y = jnp.real(y)
    # scale it up
    y = y * (N / n)
    return y


def vec_convolve(x, h):
    """1D full convolution based on a hack suggested by Jake Vanderplas

    See https://github.com/google/jax/discussions/7961 for details
    """
    return signal.convolve(x[None], h[None])[0]

vec_convolve_jit = jit(vec_convolve)




# Copyright 2022-Present CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Fast Walsh Hadamard Transforms
"""

from jax import lax, jit
import jax.numpy as jnp



@jit
def fwht(X):
    """Computes the Fast Walsh Hadamard Transform over columns

    Args:
        X (jax.numpy.ndarray): The 1D real signal or 2D matrix where each column is a signal whose transform is to be computed

    Returns:
        jax.numpy.ndarray: The Fast Walsh Hadamard Transform coefficients of (columns of) X 
    """
    n = X.shape[0]
    # number of stages
    s = (n-1).bit_length()

    def init1():
        Y = jnp.empty(X.shape, dtype=X.dtype)
        A  = X[0::2]
        B = X[1::2]
        Y = Y.at[0::2].set(A + B)
        Y = Y.at[1::2].set(A - B)
        return (Y, 1, 2, 4)

    def body1(state):
        # gap between x entries
        # number of x entries
        X, count, gap, step = state
        Y = jnp.empty(X.shape, dtype=X.dtype)
        J = 0
        k = 0
        def body2(state):
            Y, J, k = state
            def body3(state):
                Y, j, k = state
                # compute the four parts
                a = X[j]
                b = X[j+gap]
                c = X[j+1]
                d = X[j+1+gap]
                Y = Y.at[k].set(a+b)
                Y = Y.at[k+1].set(a-b)
                Y = Y.at[k+2].set(c-d)
                Y = Y.at[k+3].set(c+d)
                return (Y, j+2, k+4)
            def cond3(state):
                j = state[1]
                return j <  J+gap-1
            # the loop
            init3 = (Y, J, k)
            Y, j, k = lax.while_loop(cond3, body3, init3)
            return (Y, J + step, k)

        def cond2(state):
            k = state[2]
            return k < n - 1

        init2 = Y, J, 0
        Y, J, k = lax.while_loop(cond2, body2, init2)

        return (Y, count+1, 2*gap, 2*step)

    def cond1(state):
        count = state[1]
        return count < s

    state = lax.while_loop(cond1, body1, init1())
    return state[0]



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Algorithms and utilities for solving SVD problems
"""

from cr.nimble._src.svdpack.reorth import (
    reorth_mgs,
    reorth_mgs_jit,
    reorth_noop
)

from cr.nimble._src.svdpack.bdsqr import (
    bdsqr,
    bdsqr_jit
)

from cr.nimble._src.svdpack.lanbpro_utils import (
    LanBDOptions,
    LanBProState,
    lanbpro_options_init,
    do_elr,
    lanbpro_random_start,
    update_nu,
    update_mu,
    compute_ind,
    bpro_norm_estimate,
)

from cr.nimble._src.svdpack.lanbpro import (
    lanbpro_init,
    lanbpro_iteration,
    lanbpro_iteration_jit,
    lanbpro,
    lanbpro_jit,
    new_r_vec,
    new_p_vec,
)
from cr.nimble._src.svdpack.lansvd_utils import (
    refine_bounds,
)

from cr.nimble._src.svdpack.lansvd import (
    lansvd_simple,
    lansvd_simple_jit
)



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Signal Processing Utilities
"""

# pylint: disable=W0611


from cr.nimble._src.signal import (
    # energy of a signal
    energy,
    # interpolate via fourier transform
    interpft,
    # statistical normalization of data
    normalize,
    normalize_jit,
    # convolution
    vec_convolve,
    vec_convolve_jit,
)

from cr.nimble._src.signal import (

    find_first_signal_with_energy_le_rw,
    find_first_signal_with_energy_le_cw,
)

from cr.nimble._src.signal import (
    frequency_spectrum,
    power_spectrum
)


from cr.nimble._src.signalcomparison import (
    SignalsComparison,
    snrs_cw,
    snrs_rw,
    snr
)

from cr.nimble._src.noise import (
    awgn_at_snr
)


from cr.nimble._src.dsp.dct import (
    dct,
    idct,
    orthonormal_dct,
    orthonormal_idct
)


from cr.nimble._src.dsp.wht import (
    fwht,
)

from cr.nimble._src.dsp.synthetic_signals import (
    time_values,
)

from cr.nimble._src.dsp.util import (
    norm_freq,
)



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax.numpy as jnp
from jax.numpy.linalg import norm
from jax import jit

@jit
def factor_mgs(A):
    n, m = A.shape
    # n rows
    # m dimension
    if n > m:
        raise Exception("Number of rows is larger than dimension")
    Q = jnp.empty([n, m])
    R = jnp.zeros([n, n])
    for k in range(0, n):
        # fill the k-th diagonal entry in R
        atom = A[k]
        norm_a = norm(atom)
        R = R.at[k, k].set(norm_a)
        # Initialize the k-th vector in Q
        q = atom / norm_a
        Q = Q.at[k].set(q)
        # Compute the inner product of new q vector with each of the remaining rows in A
        products = A[k+1:n, :] @ q.T
        # Place in k-th column of R
        R = R.at[k+1:n, k].set(products)
        # Subtract the contribution of previous q vector from all remaining rows of A.
        rr = R[k+1:n, k:k+1]
        update =  -rr @ jnp.expand_dims(q, 0)
        A = A.at[k+1:n].add(update)
    return R, Q


def update(R, Q, a, k):
    if k > 0:
        # make it a column vector
        b = jnp.expand_dims(a, 1)
        # Compute the projection of a on each of the previous rows in Q
        h = Q[:k, :] @ b
        # Store in the k-th row of R
        R = R.at[k, :k].set(jnp.squeeze(h))
        # subtract the projections
        proj = h.T @ Q[:k, :]
        a = a - jnp.squeeze(proj)
    # compute norm
    a_norm = norm(a)
    # save it in the diagonal entry of R
    R = R.at[k,k].set(a_norm)
    # place the new normalized vector
    a = jnp.squeeze(a)
    a = a /a_norm
    Q = Q.at[k, :].set(a)
    return R, Q

update = jit(update, static_argnums=(3,))


def solve(R, Q, x):
    pass


# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from jax import lax, jit, vmap, random
import jax.numpy as jnp
from jax.numpy.linalg import norm

from cr.nimble import promote_arg_dtypes

def reorth_mgs(Q, r, r_norm, indices, alpha=0.5):
    """Reorthogonalizes r against subset of columns in Q indexed by indices

    If norm of r reduces significantly, then a second reorthogonalization is performed.
    If the norm of r reduces again significantly, then it is assumed that r is 
    numerically in the column span of Q and a zero vector is returned.
    """
    n, k = Q.shape
    k2 = len(indices)
    assert k2 <= k
    Q, r, r_norm = promote_arg_dtypes(Q, r, r_norm)

    def orthogonalize_with_col(i, r):
        # pick the corresponding column
        q = Q[:, i]
        # compute the dot product
        t = jnp.dot(q, r)
        # subtract the projection from r
        r = r - t * q
        return r

    def for_body(i, r):
        # orthogonalize against a column only if it is selected
        return lax.cond(indices[i], 
            lambda r : orthogonalize_with_col(i, r),
            lambda r : r,
            r
        )
    # orthogonalize r against Q
    r = lax.fori_loop(0, k2, for_body, r)
    old_norm = r_norm
    r_norm = norm(r)

    def while_cond(state):
        r, r_norm, old_norm, iterations = state
        return r_norm < alpha * old_norm

    def while_body(state):
        r, r_norm, old_norm, iterations = state
        # orthogonalize r against Q
        r = lax.fori_loop(0, k2, for_body, r)
        old_norm = r_norm
        r_norm = norm(r)
        return r, r_norm, old_norm, iterations+1

    state = r, r_norm, old_norm, 1
    state = lax.while_loop(while_cond, while_body, state)
    r, r_norm, old_norm, iterations = state
    return r, r_norm, iterations

reorth_mgs_jit = jit(reorth_mgs)

def reorth_noop(r, r_norm):
    return r, r_norm, 0



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import partial
from jax import jit, lax
import jax.numpy as jnp

from cr.nimble import promote_arg_dtypes, check_shapes_are_equal, dtype_ranges

@jit
def mean_squared(array):
    """Returns the mean squared value of an array
    """
    # We need to handle both real and complex cases
    sqr = jnp.conj(array) * array
    mean_sqr = jnp.mean(sqr)
    # Make sure that we are down to float data type
    return jnp.abs(mean_sqr)

@jit
def mean_squared_error(array1, array2):
    """Returns the mean square error between two arrays
    """
    # check shape compatibility
    check_shapes_are_equal(array1, array2)
    # promote to same inexact type (real or complex)
    array1, array2 = promote_arg_dtypes(array1, array2)
    diff = array1 - array2
    # We need to handle both real and complex cases
    sqr_error = jnp.conj(diff) * diff
    mse = jnp.mean(sqr_error)
    # Make sure that we are down to float data type
    return jnp.abs(mse)

def root_mean_squared(array):
    """Returns the root mean squared value of an array
    """
    return jnp.sqrt(mean_squared(array))


def root_mse(array1, array2):
    """Returns the root mean square error between two arrays
    """
    return jnp.sqrt(mean_squared_error(array1, array2))


@partial(jit, static_argnums=(1,))
def normalization_factor(array, normalization : str):
    """Returns the normalization factor based on the contents of an array
    """
    normalization = normalization.lower()
    if normalization == 'euclidean':
        return root_mean_squared(array)
    elif normalization == 'min-max':
        return jnp.abs(jnp.max(array) - jnp.min(array))
    elif normalization == 'mean':
        return jnp.abs(jnp.mean(array))
    elif normalization == 'median':
        return jnp.abs(jnp.median(array))
    else:
        raise ValueError("Unsupported normalization type")


@partial(jit, static_argnames=("normalization",))
def normalized_root_mse(reference_arr, test_arr, normalization='euclidean'):
    """Returns the normalized root mean square error between two arrays
    """
    rmse = root_mse(reference_arr, test_arr)
    denom = normalization_factor(reference_arr, normalization)
    # make sure that denom is non-zero
    eps = jnp.finfo(float).eps
    denom = denom + eps
    return rmse / denom

@jit
def peak_signal_noise_ratio(reference_arr, test_arr):
    """Returns the Peak Signal to Noie Ratio between two arrays 
    """
    min_val, max_val = dtype_ranges[reference_arr.dtype]
    data_min = jnp.min(reference_arr)
    zero = jnp.zeros_like(min_val)
    # min_val below 0 is considered only if the data really has negative values
    min_val = jnp.where(data_min >= 0, zero, min_val)
    drange = max_val - min_val
    mse = mean_squared_error(reference_arr, test_arr)
    eps = jnp.finfo(float).eps
    mse = lax.cond(mse, lambda _ : mse, lambda _ : eps, None)
    return 10 * jnp.log10((drange ** 2) / mse)

@jit
def signal_noise_ratio(reference_arr, test_arr):
    """Returns the signal to noise ratio between a reference array and a test array

    Args:
        reference_arr (jax.numpy.ndarray): Reference signal (can be ND array)
        test_arr (jax.numpy.ndarray): Test signal (can be ND array)

    Returns:
        Signal to Noise Ratio between reference and test signals

    Example:
        ::

            >>> x = jnp.ones(10)
            >>> y = 0.9 * x
            >>> signal_noise_ratio(x, y)
            DeviceArray(19.999998, dtype=float32)
    """
    reference_arr, test_arr = promote_arg_dtypes(reference_arr, test_arr)
    ref_energy = jnp.abs(jnp.vdot(reference_arr, reference_arr))
    error = reference_arr - test_arr
    err_energy = jnp.abs(jnp.vdot(error, error))
    eps = jnp.finfo(float).eps
    # make sure that error energy is non-zero
    err_energy = lax.cond(err_energy, lambda _ : err_energy, lambda _ : eps, None)
    # make sure that ref energy is non-zero
    ref_energy = lax.cond(ref_energy, lambda _ : ref_energy, lambda _ : eps, None)
    return 10 * jnp.log10(ref_energy/ err_energy)




# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from jax import lax, jit, vmap, random
import jax.numpy as jnp
from jax.numpy.linalg import norm


import cr.nimble as cnb
from cr.nimble import AH_v

from .reorth import reorth_mgs
from .lanbpro_utils import (
    LanBDOptions,
    LanBProState,
    lanbpro_options_init,
    do_elr,
    update_mu,
    update_nu,
    compute_ind,
    bpro_norm_estimate
)

FUDGE = 1.01
M2 = 3/2
N2 = 3/2

KEY = random.PRNGKey(0)
KEYS = random.split(KEY, 20)


def lanbpro_init(A, k, p0, options: LanBDOptions):
    """Initialize the state with a starting vector 
    """
    # we follow steps from page 30 of Larsen paper
    m, n = A.shape
    U = jnp.zeros((m, k))
    V = jnp.zeros((n, k))
    alpha = jnp.zeros(k)
    beta = jnp.zeros(k+1)
    mu = jnp.zeros(k)
    nu = jnp.zeros(k)
    mumax = jnp.zeros(k)
    numax = jnp.zeros(k)
    indices = jnp.zeros(k, dtype=bool)
    anorms = jnp.zeros(k)
    # options 
    delta = options.delta
    eps = options.eps
    gamma = options.gamma
    # whether full reorthogonalization will be done
    b_fro = delta == 0
    # initial value of force reorthogonalization
    b_force_reorth = False
    # step 1
    p_norm = norm(p0)
    # beta_0
    beta = beta.at[0].set(p_norm)
    # U_0
    u = cnb.vec_safe_divide_by_scalar(p0, p_norm)
    U = U.at[:, 0].set(u)
    # step 2 r update
    r = AH_v(A, u)
    # step 2.1b alpha update
    r_norm = norm(r)
    alpha = alpha.at[0].set(r_norm)
    anorm = FUDGE * r_norm
    # step 2.1b v update
    v = cnb.vec_safe_divide_by_scalar(r, r_norm)
    V = V.at[:, 0].set(v)
    # step 2.1b p update
    p = A @ v  - alpha[0] * u
    # step 2.2b beta update
    p_norm = norm(p)
    p, p_norm, p_proj = do_elr(u, p, p_norm, options.gamma)
    # Check for convergence or failure to maintain semiorthogonality
    semiorth_cond = p_norm < max(m, n) * anorm * eps
    p, p_norm, b_force_reorth, indices = lax.cond(semiorth_cond, 
        # compute a new random p vector orthogonal to previous U
        lambda _ : (new_p_vec(A, U, 1, gamma)[0], 0., True, indices.at[0].set(True)),
        lambda _ : (p, p_norm, b_force_reorth, indices), 
        None) 
    beta = beta.at[1].set(p_norm)
    # update anorm estimate 
    anorm = jnp.maximum(anorm,FUDGE*jnp.hypot(alpha[0],beta[1]))
    # update mu for the first iteration before computation of U_1
    eps = jnp.finfo(float).eps
    eps1 = 100*eps/2
    T = eps1*(anorm + jnp.hypot(alpha[0],beta[1]) + jnp.hypot(alpha[0],beta[0]) )
    # TODO this is problematic if p_norm is 0
    mu0 = lax.cond(p_norm, lambda _ :  T / p_norm, lambda _ : 0., None)
    mu = mu.at[0].set(mu0)
    # mumax update
    mumax = mumax.at[0].set(jnp.abs(mu[0]))
    # TODO add condition with elr > 0
    mu = mu.at[0].set(M2 * eps)
    # prepare state after completion of one iteration
    anorms = anorms.at[0].set(anorm)
    return  LanBProState(p=p, U=U, V=V, alpha=alpha, beta=beta,
        mu=mu, nu=nu, mumax=mumax, numax=numax,
        anorm=anorm, anorms=anorms,
        indices=indices, 
        b_force_reorth=b_force_reorth, b_fro=b_fro, iterations=1,
        )


def lanbpro_iteration(A, state: LanBProState, options: LanBDOptions):
    """One single (j-th) iteration of Lanczos bidiagonalization with partial reorthogonalization algorithm
    """
    m, n = A.shape
    max_m_n = max(m, n)
    # copy variables from the state
    p = state.p
    U = state.U
    V = state.V
    alpha = state.alpha
    beta = state.beta
    mu = state.mu
    nu = state.nu
    mumax = state.mumax
    numax = state.numax
    indices = state.indices
    anorm = state.anorm
    b_fro = state.b_fro
    b_force_reorth = state.b_force_reorth
    b_est_anorm = state.b_est_anorm
    # the total number of iterations
    k = len(alpha)
    # index for k
    idx = jnp.arange(k)
    # iteration number 
    j = state.iterations
    # first j indices mask
    j_mask =  idx < j
    jp1_mask = idx <= j
    # options
    gamma = options.gamma
    elr = options.elr
    eps = options.eps
    delta = options.delta
    eta = options.eta
    # carry out the work for one iteration of lanbpro
    beta_j = beta[j]
    # compute next left singular vector
    u = cnb.vec_safe_divide_by_scalar(p, beta_j)
    U = U.at[:, j].set(u)
    # once sufficient iterations are completed, we can
    # compute a better estimate of a norm
    anorm, b_est_anorm = lax.cond(j == 5, 
        # Replace norm estimate with largest Ritz value.
        lambda _ : (FUDGE*bpro_norm_estimate(alpha, beta), False),
        # continue with current value
        lambda _ : (anorm, b_est_anorm),
        None
    )
    # step 2 r update
    v_jm1 = V[:, j-1]
    r = AH_v(A, u) - beta_j * v_jm1
    r_norm = norm(r)
    # elr condition
    b_no_fro = jnp.logical_not(b_fro)
    elr_cond = jnp.logical_and(jnp.logical_and(r_norm < gamma * beta_j, elr), b_no_fro)
    # extended local reorthogonalization of r w.r.t. previous v_j
    r, r_norm, proj = lax.cond(elr_cond, 
        lambda _ : do_elr(v_jm1, r, r_norm, gamma),
        lambda _ : (r, r_norm, 0.),
        None
    )
    # save updated r_norm in alpha_j
    alpha = alpha.at[j].set(r_norm)
    # make changes to beta_j if required.
    beta = beta.at[j].add(proj)
    # norm estimate
    anorm_up_1 = lambda anorm: jnp.maximum(anorm,
        FUDGE*jnp.sqrt(alpha[0]**2+beta[1]**2+alpha[1]*beta[1]))
    anorm_up_j = lambda anorm: jnp.maximum(anorm,
        FUDGE*jnp.sqrt(alpha[j-1]**2+beta[j]**2
            + alpha[j-1]*beta[j-1] + alpha[j]*beta[j]))
    anorm = lax.cond(b_est_anorm,
        # We need to update norm estimate
        lambda anorm : lax.cond(j == 1, anorm_up_1, anorm_up_j, anorm),
        # no more norm estimation needed
        lambda anorm : anorm,
        anorm)
    # nu update condition
    nu_update_cond = jnp.logical_and(b_no_fro, r_norm != 0)
    nu, numax = lax.cond(nu_update_cond, 
        lambda nu: update_nu(nu, numax, mu, j, alpha, beta, anorm),
        lambda nu : (nu, numax),
        nu
    )
    # if elr is on, then current vector is orthogonalized against previous one
    nu = lax.cond(elr > 0,
        lambda nu: nu.at[j-1].set(N2 * eps),
        lambda nu: nu,
        nu)
    # condition for partial or full reorthogonalization
    reorth_cond = jnp.logical_or(b_fro, numax[j] > delta)
    reorth_cond = jnp.logical_or(reorth_cond, b_force_reorth)
    reorth_cond = jnp.logical_and(reorth_cond, alpha[j] != 0)
    # function to reorth r w.r.t. previous V
    def reorth_r(_):
        # identify the indices at which partial or full reorthogonalization will be done
        reorth_indices = lax.cond(jnp.logical_or(b_fro, eta == 0),
            # full reorthogonalization case
            lambda _ : j_mask,
            # partial reorthogonalization case
            lambda _ : lax.cond(b_force_reorth,
                lambda _ : indices,
                lambda _ : compute_ind(nu, delta, eta),
                None
            ),
            None
        )
        # carry out reorthogonalization
        r2, r2_norm, iters = reorth_mgs(V, r, r_norm, reorth_indices, gamma)
        # reset nu in the entries which have been reorthogonalized
        nu2 = jnp.where(reorth_indices, N2*eps, nu)
        # if a reorthogonalization was forced. it won't be for next iteration
        b_force_reorth2 = jnp.logical_not(b_force_reorth)
        return r2, r2_norm, reorth_indices, nu2, b_force_reorth2
    # reorthogonalize r if required
    r, r_norm, indices, nu, b_force_reorth = lax.cond(reorth_cond,
        reorth_r,
        lambda _ : (r, r_norm, indices, nu, b_force_reorth),
        None
    )
    # Check for convergence or failure to maintain semiorthogonality
    # this is the case where r is in the column space of previous V vectors
    semiorth_cond = r_norm < max(m, n) * anorm * eps
    r, r_norm, b_force_reorth, indices = lax.cond(semiorth_cond, 
        # compute a new random r vector orthogonal to previous V
        lambda _ : (new_r_vec(A, V, j, gamma)[0], 0., True, j_mask),
        lambda _ : (r, r_norm, b_force_reorth, indices), 
        None) 
    # update alpha_j again if required
    alpha = alpha.at[j].set(r_norm)
    # step 2.1b v update
    v = cnb.vec_safe_divide_by_scalar(r, r_norm)
    V = V.at[:, j].set(v)
    # Lanczos step to generate u_{j+1}
    # step 2.1b p update
    p = A @ v  - alpha[j] * u
    # step 2.2b beta update
    p_norm = norm(p)
    # elr condition for p
    elr_cond = jnp.logical_and(jnp.logical_and(p_norm < gamma * r_norm, elr), b_no_fro)
    # extended local reorthogonalization of p w.r.t. previous u_j
    p, p_norm, proj = lax.cond(elr_cond, 
        lambda _ : do_elr(u, p, p_norm, gamma),
        lambda _ :(p, p_norm, 0.),
        None
    )
    # save updated p_norm in beta_{j+1} (if there are any changes)
    beta = beta.at[j+1].set(p_norm)
    # make changes to alpha_j if required.
    alpha = alpha.at[j].add(proj)
    anorm = lax.cond(b_est_anorm,
        # we need to update anorm estimate
        lambda anorm: jnp.maximum(anorm,FUDGE*jnp.sqrt(alpha[j]**2+beta[j+1]**2+alpha[j]*beta[j])),
        # no further need to update norm estimate
        lambda anorm: anorm,
        anorm)

    # mu update condition
    mu_update_cond = jnp.logical_and(b_no_fro, p_norm != 0)
    mu, mumax = lax.cond(mu_update_cond, 
        lambda mu: update_mu(mu, mumax, nu, j, alpha, beta, anorm),
        lambda mu : (mu, mumax),
        mu
    )
    # TODO add condition with elr > 0
    mu = mu.at[j].set(M2 * eps)
    # condition for partial or full reorthogonalization
    reorth_cond = jnp.logical_or(b_fro, mumax[j] > delta)
    reorth_cond = jnp.logical_or(reorth_cond, b_force_reorth)
    reorth_cond = jnp.logical_and(reorth_cond, p_norm != 0)
    # function to reorth p w.r.t. previous U
    def reorth_p(_):
        # identify the indices at which partial or full reorthogonalization will be done
        # from U_0 to U_j [j+1 vectors] have already been computed 
        reorth_indices = lax.cond(jnp.logical_or(b_fro, eta == 0),
            # full reorthogonalization case
            lambda _ : jp1_mask,
            # partial reorthogonalization case
            lambda _ : lax.cond(b_force_reorth,
                # for forced reorth, we need to add one more vec
                lambda _ : indices.at[k - jnp.argmax(indices[::-1])].set(True),
                lambda _ : compute_ind(mu, delta, eta),
                None
            ),
            None
        )
        # carry out reorthogonalization
        p2, p2_norm, iters = reorth_mgs(U, p, p_norm, indices, gamma)
        # reset mu in the entries which have been reorthogonalized
        mu2 = jnp.where(reorth_indices, M2*eps, mu)
        # if a reorthogonalization was forced. it won't be for next iteration
        b_force_reorth2 = jnp.logical_not(b_force_reorth)
        return p2, p2_norm, reorth_indices, mu2, b_force_reorth2
    # reorthogonalize p if required
    p, p_norm, indices, nu, b_force_reorth = lax.cond(reorth_cond,
        reorth_p,
        lambda _ : (p, p_norm, indices, nu, b_force_reorth),
        None
    )
    # Check for convergence or failure to maintain semiorthogonality
    semiorth_cond = p_norm < max(m, n) * anorm * eps
    p, p_norm, b_force_reorth, indices = lax.cond(semiorth_cond, 
        # compute a new random p vector orthogonal to previous U
        lambda _ : (new_p_vec(A, U, j+1, gamma)[0], 0., True, jp1_mask),
        lambda _ : (p, p_norm, b_force_reorth, indices), 
        None) 
    # save updated p_norm in beta_{j+1} (if there are any changes)
    beta = beta.at[j+1].set(p_norm)

    # track anorm for the current iteration
    anorms = state.anorms.at[j].set(anorm)
    # prepare the state for next iteration
    return  LanBProState(p=p, U=U, V=V, alpha=alpha, beta=beta,
        mu=mu, nu=nu, mumax=mumax, numax=numax,
        anorm=anorm, anorms=anorms,
        indices=indices, b_fro=b_fro, b_force_reorth=b_force_reorth,
        b_est_anorm=b_est_anorm, iterations=j+1
        )


lanbpro_iteration_jit = jit(lanbpro_iteration)


def lanbpro(A, k, p0):
    """K steps of the  Lanczos bidiagonalization with partial reorthogonalization
    """
    options = lanbpro_options_init(k)
    state  = lanbpro_init(A, k, p0, options)

    def cond(state):
        return state.iterations < k

    def body(state):
        state =  lanbpro_iteration(A, state, options, state.iterations)
        return state

    # state = lax.while_loop(cond, body, state)
    # while cond(state):
    #     state = body(state)
    state = lax.fori_loop(1, k, 
        lambda i, state: lanbpro_iteration(A, state, options),
        state)
    return state

lanbpro_jit = jit(lanbpro, static_argnums=(1,))


def new_p_vec(A, U, j,  gamma):
    """Generates a new p vector which is orthogonal to all previous U vectors
    """
    m, n = A.shape
    idx = jnp.arange(U.shape[1])
    j_mask =  idx < j

    def p_vec(i):
        p = random.uniform(KEYS[i], (n,))
        p = A @ p
        p_norm = norm(p)
        p, p_norm, iters = reorth_mgs(U, p, p_norm, j_mask, gamma)
        return p, p_norm

    def init():
        p, p_norm = p_vec(0)
        return p, p_norm, 1

    def cond(state):
        p, p_norm, iterations = state
        cond = jnp.logical_and(iterations < 10, p_norm <1e-10)
        return cond

    def body(state):
        p, p_norm, i = state
        p, p_norm  = p_vec(i)
        return p, p_norm, i+1

    state = lax.while_loop(cond, body, init())
    p, p_norm, i = state
    return p / p_norm, i

new_p_vec_jit = jit(new_p_vec)


def new_r_vec(A, V, j,  gamma):
    """Generates a new r vector which is orthogonal to all previous V vectors
    """
    m, n = A.shape
    idx = jnp.arange(V.shape[1])
    j_mask =  idx < j
    max_iters = 20

    def r_vec(i):
        # This approach fails for large matrices with low rank
        # r = random.uniform(keys[i], (m,))
        # r = AH_v(A, r)
        r = random.uniform(KEYS[i], (n,))
        r_norm = norm(r)
        r, r_norm, iters = reorth_mgs(V, r, r_norm, j_mask, gamma)
        return r, r_norm

    def init():
        r, r_norm = r_vec(0)
        return r, r_norm, 1

    def cond(state):
        r, r_norm, iterations = state
        cond = jnp.logical_and(iterations < max_iters, r_norm <1e-10)
        return cond

    def body(state):
        r, r_norm, i = state
        r, r_norm  = r_vec(i)
        return r, r_norm, i+1

    state = lax.while_loop(cond, body, init())
    r, r_norm, i = state
    # print(f'r_norm {r_norm}, iters: {i}')
    return r / r_norm, i

new_r_vec_jit = jit(new_r_vec)



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp

def hermitian(A):
    r"""Returns the Hermitian transpose of an array

    Args:
        A (jax.numpy.ndarray): An array

    Returns:
        (jax.numpy.ndarray): An array: :math:`A^H`
    """
    return jnp.conjugate(A.T)

def check_shapes_are_equal(array1, array2):
    """Raise an error if the shapes of the two arrays do not match.
    
    Raises:
        ValueError: if the shape of two arrays is not same
    """
    if not array1.shape == array2.shape:
        raise ValueError('Input arrays must have the same shape.')
    return



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for solving dense linear systems
"""

from jax.scipy.linalg import solve
from jax.numpy.linalg import lstsq

def mult_with_submatrix(A, columns, x):
    """Computes :math:`b = A[:, I] x`
    """
    A = A[:, columns]
    return A @ x


def solve_on_submatrix(A, columns, b):
    """Solves the problem :math:`A[:, I] x = b` where I is an
    index set of selected columns
    """
    A = A[:, columns]
    x, r_norms, rank, s = lstsq(A, b)
    r = b - A @ x
    return x, r




# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import jax.numpy as jnp

import jax
from jax import lax, random
from jax._src import dtypes

from jax.lib import xla_bridge
platform = xla_bridge.get_backend().platform


def is_cpu():
    """Returns True if the code is running on a CPU platform
    """
    return platform == 'cpu'

def is_gpu():
    """Returns True if the code is running on a GPU platform
    """
    return platform == 'gpu'

def is_tpu():
    """Returns True if the code is running on a TPU platform
    """
    return platform == 'tpu'

KEY0 = random.PRNGKey(0)
KEYS = random.split(KEY0, 64)

def promote_arg_dtypes(*args):
    """Promotes `args` to a common inexact type.
    
    Args:
        *args: list of JAX ndarrays to be promoted to common inexact type

    Returns:
        The same list of arrays with their dtype promoted to a common inexact type

    Example:
        Promoting a single argument::

            >>> cr.nimble.promote_arg_dtypes(jnp.arange(2))
            DeviceArray([0., 1.], dtype=float32)
            >>> from jax.config import config
            >>> config.update("jax_enable_x64", True)
            >>> cr.nimble.promote_arg_dtypes(jnp.arange(2))
            DeviceArray([0., 1.], dtype=float64)

        Promoting two arguments to common floating point type::

            >>> a = jnp.arange(2)
            >>> b = jnp.arange(1.5, 3.5)
            >>> a, b = cr.nimble.promote_arg_dtypes(a, b)
            >>> print(a)
            >>> print(b)
            [0. 1.]
            [1.5 2.5]

        A mix of real and complex types::

            >>> a = jnp.arange(2) + 0.j
            >>> b = jnp.arange(1.5, 3.5)
            >>> a, b = cr.nimble.promote_arg_dtypes(a, b)
            >>> print(a)
            >>> print(b)
            [0.+0.j 1.+0.j]
            [1.5+0.j 2.5+0.j]
    """
    def _to_inexact_type(type):
        return type if jnp.issubdtype(type, jnp.inexact) else jnp.float_
    inexact_types = [_to_inexact_type(arg.dtype) for arg in args]
    dtype = dtypes.canonicalize_dtype(jnp.result_type(*inexact_types))
    args = [lax.convert_element_type(arg, dtype) for arg in args]
    if len(args) == 1:
        return args[0]
    else:
        return args


def canonicalize_dtype(dtype):
    """Wrapper function on dtypes.canonicalize_dtype with None handling
    """
    if dtype is None:
        return dtype
    return dtypes.canonicalize_dtype(dtype)


def promote_to_complex(arg):
    """Promotes an argument to complex type"""
    dtype = dtypes.result_type(arg, np.complex64)
    return lax.convert_element_type(arg, dtype)

def promote_to_real(arg):
    """Promotes an argument to real type"""
    dtype = dtypes.result_type(arg, np.float32)
    return lax.convert_element_type(arg, dtype)


# Integer types
integer_types = (
    jnp.uint8.dtype,
    jnp.uint16.dtype,
    jnp.uint32.dtype,
    jnp.uint64.dtype,
    jnp.int8.dtype,
    jnp.int16.dtype,
    jnp.int32.dtype,
    jnp.int64.dtype,
)

# Ranges of values for integer types
integer_ranges = {t: (jnp.iinfo(t).min, jnp.iinfo(t).max)
                   for t in integer_types}

# Ranges of values for floating point types
dtype_ranges = {
    bool: (False, True),
    float: (-1, 1),
    jnp.bool_.dtype: (False, True),
    jnp.float_.dtype: (-1, 1),
    jnp.float16.dtype: (-1, 1),
    jnp.float32.dtype: (-1, 1),
    jnp.complex64.dtype: (-1, 1),
    jnp.complex128.dtype: (-1, 1),
}

dtype_ranges.update(integer_ranges)


def nbytes_live_buffers():
    """Returns the number of bytes consumed by the live buffers
    """
    backend = jax.lib.xla_bridge.get_backend()
    nbytes = [buf.nbytes for buf in backend.live_buffers()]
    return np.sum(nbytes)



"""
Utility functions for working with vectors
"""

from jax import jit, lax
import jax.numpy as jnp

from .util import promote_arg_dtypes

def is_scalar(x):
    """Returns if x is a scalar 

    Args:
        x (jax.numpy.ndarray): A JAX array.

    Returns:
        True if x is a scalar quantity (i.e. ndim==0).
    """
    return x.ndim == 0

def is_vec(x):
    """Returns if x is a line vector or row vector or column vector

    Args:
        x (jax.numpy.ndarray): A JAX array.

    Returns:
        True if x is a line vector or a row vector or a column vector.
    """
    return x.ndim == 1 or (x.ndim == 2 and 
        (x.shape[0] == 1 or x.shape[1] == 1))

def is_line_vec(x):
    """Returns if x is a line vector

    Args:
        x (jax.numpy.ndarray): A JAX array.

    Returns:
        True if x is a line vector.
    """
    return x.ndim == 1

def is_row_vec(x):
    """Returns if x is a row vector

    Args:
        x (jax.numpy.ndarray): A JAX array.

    Returns:
        True if x is a row vector.
    """
    return x.ndim == 2 and x.shape[0] == 1 

def is_col_vec(x):
    """Returns if x is a column vector

    Args:
        x (jax.numpy.ndarray): A JAX array.

    Returns:
        True if x is a column vector.
    """
    return x.ndim == 1 or (x.ndim == 2 and x.shape[1] == 1)

def to_row_vec(x):
    """Converts a line vector to a row vector

    Args:
        x (jax.numpy.ndarray): A line vector (ndim == 1).

    Returns:
        jax.numpy.ndarray: A row vector.
    """
    assert x.ndim == 1
    return jnp.expand_dims(x, 0)

def to_col_vec(x):
    """Converts a line vector to a column vector

    Args:
        x (jax.numpy.ndarray): A line vector (ndim == 1).

    Returns:
        jax.numpy.ndarray: A column vector.
    """
    assert x.ndim == 1
    return jnp.expand_dims(x, 1)

def vec_unit(n, i):
    """Returns a unit vector in i-th dimension for the standard coordinate system

    Args:
        n (int): Length of the vector.
        i (int): Index/dimension of the unit vector.

    Returns:
        jax.numpy.ndarray: A line vector of length n with all zeros except a one at position i. 
    """
    return jnp.zeros(n).at[i].set(1)

vec_unit_jit = jit(vec_unit, static_argnums=(0, 1))

def vec_shift_right(x):
    """Right shift the contents of the vector

    Args:
        x (jax.numpy.ndarray): A line vector.

    Returns:
        jax.numpy.ndarray: Right shifted x. 
    """
    return jnp.zeros_like(x).at[1:].set(x[:-1])

def vec_rotate_right(x):
    """Circular right shift the contents of the vector

    Args:
        x (jax.numpy.ndarray): A line vector.

    Returns:
        jax.numpy.ndarray: Right rotated x. 
    """
    return jnp.roll(x, 1)


def vec_shift_left(x):
    """Left shift the contents of the vector

    Args:
        x (jax.numpy.ndarray): A line vector.

    Returns:
        jax.numpy.ndarray: Left shifted x. 
    """
    return jnp.zeros_like(x).at[0:-1].set(x[1:])

def vec_rotate_left(x):
    """Circular left shift the contents of the vector

    Args:
        x (jax.numpy.ndarray): A line vector.

    Returns:
        jax.numpy.ndarray: Left rotated x. 
    """
    return jnp.roll(x, -1)

def vec_shift_right_n(x, n):
    """Right shift the contents of the vector by n places

    Args:
        x (jax.numpy.ndarray): A line vector.
        n (int): Number of positions to shift.

    Returns:
        jax.numpy.ndarray: Right shifted x by n places. 
    """
    return jnp.zeros_like(x).at[n:].set(x[:-n])

def vec_rotate_right_n(x, n):
    """Circular right shift the contents of the vector by n places

    Args:
        x (jax.numpy.ndarray): A line vector.
        n (int): Number of positions to shift.

    Returns:
        jax.numpy.ndarray: Right roted x by n places. 
    """
    return jnp.roll(x, n)


def vec_shift_left_n(x, n):
    """Left shift the contents of the vector by n places

    Args:
        x (jax.numpy.ndarray): A line vector.
        n (int): Number of positions to shift.

    Returns:
        jax.numpy.ndarray: Left shifted x by n places. 
    """
    return jnp.zeros_like(x).at[0:-n].set(x[n:])

def vec_rotate_left_n(x, n):
    """Circular left shift the contents of the vector by n places

    Args:
        x (jax.numpy.ndarray): A line vector.
        n (int): Number of positions to shift.

    Returns:
        jax.numpy.ndarray: Left rotated x by n places. 
    """
    return jnp.roll(x, -n)

def vec_safe_divide_by_scalar(x, alpha):
    return lax.cond(alpha == 0, lambda x : x, lambda x: x / alpha, x)

vec_safe_divide_by_scalar_jit = jit(vec_safe_divide_by_scalar)


def vec_repeat_at_end(x, p):
    """Extends a vector by repeating it at the end (periodic extension)

    Args:
        x (jax.numpy.ndarray): A line vector.
        p (int): Number of samples by which x will be extended.

    Returns:
        jax.numpy.ndarray: x extended periodically at the end. 
    """
    n = x.shape[0]
    indices = jnp.arange(p) % n
    padding = x[indices]
    return jnp.concatenate((x, padding))

vec_repeat_at_end_jit = jit(vec_repeat_at_end, static_argnums=(1,))


def vec_repeat_at_start(x, p):
    """Extends a vector by repeating it at the start (periodic extension)

    Args:
        x (jax.numpy.ndarray): A line vector.
        p (int): Number of samples by which x will be extended.

    Returns:
        jax.numpy.ndarray: x extended periodically at the start. 
    """
    n = x.shape[0]
    indices = (jnp.arange(p) + n - p) % n
    padding = x[indices]
    return jnp.concatenate((padding, x))

vec_repeat_at_start_jit = jit(vec_repeat_at_start, static_argnums=(1,))


def vec_centered(x, length):
    """Returns the central part of a vector of a specified length

    Args:
        x (jax.numpy.ndarray): A line vector.
        length (int): Length of the central part of x which will be retained.

    Returns:
        jax.numpy.ndarray: central part of x of the specified length. 
    """
    cur_len = len(x)
    length = min(cur_len, length) 
    start = (len(x) - length) // 2
    end = start + length
    return x[start:end]

vec_centered_jit = jit(vec_centered, static_argnums=(1,))

########################################################
#  Energy
########################################################

@jit
def vec_mag_desc(a):
    """Returns the coefficients in the descending order of magnitude

    Args:
        a (jax.numpy.ndarray): A vector of coefficients
    """
    return jnp.sort(jnp.abs(a))[::-1]

@jit
def vec_to_pmf(a):
    """Computes a probability mass function from a given vector

    Args:
        a (jax.numpy.ndarray): A vector of coefficients
    """
    s = jnp.sum(a) * 1.
    return a / s

@jit
def vec_to_cmf(a):
    """Computes a cumulative mass function from a given vector

    Args:
        a (jax.numpy.ndarray): A vector of coefficients
    """
    s = jnp.sum(a) * 1.
    # normalize
    a = a / s
    # generate the CMF
    return jnp.cumsum(a)

@jit
def cmf_find_quantile_index(a, q):
    """Returns the index of a given quantile in a CMF

    Args:
        a (jax.numpy.ndarray): A vector of coefficients
    """
    return jnp.argmax(a >= q)

def num_largest_coeffs_for_energy_percent(a, p):
    """Returns the number of largest components containing a given
    percentage of energy

    Args:
        a (jax.numpy.ndarray): A vector of coefficients
        p (float): percentage of energy
    """
    # compute energies
    a = a ** 2
    # sort in descending order
    a = jnp.sort(a)[::-1]
    # total energy
    s = jnp.sum(a) * 1.
    # normalize
    a = a / s
    # convert to a cmf
    cmf = jnp.cumsum(a)
    # the quantile value
    q = p / 100.
    # find the index
    index =  jnp.argmax(cmf >= q)
    return index + 1

########################################################
#  Sliding Windows
########################################################

def vec_to_windows(x, wlen):
    """Constructs windows of a given length from the vector

    Args:
        x (jax.numpy.ndarray): A line vector
        wlen: length of each window

    Returns:
        jax.numpy.ndarray: A matrix of shape (wlen, m) where
        m is the number of windows

    Notes:
    - Drops extra samples from the end if the last window is not complete
    """
    n = len(x)
    # number of windows
    m = n // wlen
    # total samples to be kept
    s = m * wlen
    return jnp.reshape(x[:s], (m, wlen)).T

vec_to_windows_jit = jit(vec_to_windows, static_argnums=(1,))



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from jax import lax, jit, vmap, random
import jax.numpy as jnp
from jax.numpy.linalg import norm


def refine_bounds(D, bnd, tol):
    """Examins the gap structure of Ritz eigen values and refines error bounds
    """
    return bnd


# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import NamedTuple
import math

from jax import lax, jit, vmap
import jax.numpy as jnp
from jax.numpy.linalg import norm
from jax.scipy.linalg import svd

def bdsqr(alpha, beta, k):
    """Computes the SVD of the bidiagonal matrix
    """
    # prepare the k+1 x k bidiagonal matrix
    B = jnp.zeros((k+1, k))
    # diagonal indices for k alpha entries
    indices = jnp.diag_indices(k)
    B = B.at[indices].set(alpha[:k])
    # subdiagonal indices for k beta entries (from second row)
    rows, cols = indices
    rows = rows + 1
    B = B.at[(rows, cols)].set(beta[1:k+1])
    # print(B)
    # perform full svd
    U, s, Vh = svd(B, full_matrices=False, compute_uv=True)
    # print(s)
    # pick the last row of U as the bounds
    bnd = U[-1, :k]
    # print(bnd)
    return U, s, Vh, bnd

bdsqr_jit = jit(bdsqr, static_argnums=(2,))


# Copyright 2021 CR.Sparse Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import jax.numpy as jnp
from sympy.ntheory import factorint


def next_pow_of_2(n):
    """
    Returns the smallest integer greater than or equal to n which is a power of 2
    """
    return 2**int(math.ceil(math.log2(n)))

def is_integer(x):
    return jnp.mod(x, 1) == 0

def is_positive_integer(x):
    return jnp.logical_and(x > 0, jnp.mod(x, 1) == 0)

def is_negative_integer(x):
    return jnp.logical_and(x < 0, jnp.mod(x, 1) == 0)


def is_odd(x):
    return jnp.mod(x, 2) == 1

def is_even(x):
    return jnp.mod(x, 2) == 0

def is_odd_natural(x):
    return jnp.logical_and(x > 0, jnp.mod(x, 2) == 1)


def is_even_natural(x):
    return jnp.logical_and(x > 0, jnp.mod(x, 2) == 0)

def is_power_of_2(x):
    return jnp.logical_not(jnp.bitwise_and(x, x - 1))

def is_perfect_square(x):
    return is_integer(jnp.sqrt(x))


def integer_factors_close_to_sqr_root(n):
    assert isinstance(n, int)
    a_max = math.floor(math.sqrt(n))
    if n % a_max == 0:
        a = a_max
        b = n // a
        return a,b
    # get the prime factors
    factors_map = factorint(n)
    factors = factors_map.keys()
    candidates = {1}
    #print(a_max)
    for key in factors_map:
        for count in range(factors_map[key]):
            new_candidates = {key*c for c in candidates}
            candidates = candidates.union(new_candidates)
            # filter out larger candidates
            candidates = {c for c in candidates if c <= a_max}
            #print(candidates)
    # a is the last candidate
    candidates = list(candidates)
    candidates.sort()
    a = candidates[-1]
    b = n // a
    return a, b


# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Random subspaces
"""

import jax.numpy as jnp
from jax import random, jit

import cr.nimble as cnb

def random_subspaces(key, N, D, K):
    """Generates a set of orthonormal bases for random low dimensional subspaces

    Args:
        key: a PRNG key used as the random key.
        N (int): Dimension of the ambient space
        D (int): Dimension of the low dimensional subspace
        K (int): Number of low dimensional subspaces

    Returns:
        (:obj:`list` of :obj:`jax.numpy.ndarray`): A list of orthonormal bases for the random low dimensional subspaces
    """
    keys = random.split(key, K)
    bases = []
    for i in range(K):
        A = random.normal(keys[i], [N, D])
        Q, _ = jnp.linalg.qr(A)
        bases.append(Q)
    return bases

random_subspaces_jit = jit(random_subspaces, static_argnums=(1,2,3,))


def uniform_points_on_subspaces(key, bases, n):
    """Generates a set of nk points on the unit sphere of each of the subspaces

    Args:
        key: a PRNG key used as the random key.
        bases (:obj:`list` of :obj:`jax.numpy.ndarray`): List of orthonormal bases for the low dimensional subspaces
        n (int): Number of points on each subspace unit sphere

    Returns:
        (jax.numpy.ndarray): A matrix containing the list of points
    """
    # number of subspaces
    K = len(bases)
    # total number of points
    total = K * n
    # the ambient dimension
    N = bases[0].shape[0]
    # allocate the space
    X = jnp.zeros((N, total), dtype=bases[0].dtype)
    keys = random.split(key, K)
    start = 0
    for i in range(K):
        A = bases[i]
        # dimension of the subspace
        di = A.shape[1]
        # Generate coefficients for the subspace
        coeffs = random.normal(keys[i], [di, n])
        # Normalize the coefficients
        coeffs = cnb.normalize_l2_cw(coeffs)
        # Compute the points
        points = A @ coeffs
        X = X.at[:, start:start+n].set(points)
        start += n
    return X

uniform_points_on_subspaces_jit = jit(uniform_points_on_subspaces, static_argnums=(2,))

def two_subspaces_at_angle(key, N, D, theta):
    """Returns ONBs for two subspaces at angle theta with each other

    Args:
        key: a PRNG key used as the random key.
        N (int): Dimension of the ambient space
        D (int): Dimension of the low dimensional subspace
        theta (float): Smallest principal angle (in degrees) between the two subspaces

    Returns:
        (jax.numpy.ndarray, jax.numpy.ndarray): A tuple (A, B) consiting of two
        ONBs for the two subspaces

    Example:
        >>> rkey = random.PRNGKey(1)
        >>> N = 20
        >>> D = 4
        >>> theta = 15
        >>> A, B = two_subspaces_at_angle(rkey, N, D, theta)
        >>> print(A.shape, B.shape)
        (20, 4) (20, 4) (20, 4)
        >>> from cr.nimble.subspaces import principal_angles_deg
        >>> print(principal_angles_deg(A, B))
        [15. 90. 90. 90.]
    """
    # Convert theta to radians
    theta = jnp.deg2rad(theta)
    # Draw two random vectors in the ambient space
    X = random.normal(key, (N, 2))
    # Orthogonalize them
    U, s, VH = jnp.linalg.svd(X, full_matrices=False)
    # Pick these two vectors
    a1 = U[:, 0]
    a2 = U[:, 1]
    #  expected value of inner product between the vectors
    p = jnp.cos(theta)
    # linear combination terms for the first vector of the second space
    c1 = p
    s1 = jnp.sqrt(1 - c1**2)
    # first vector for second space
    b1 = s1 * a2 + c1 * a1
    # put these vectors together
    X = jnp.column_stack((a1, b1))
    # Find the orthogonal complement of X 
    U, s, VH = jnp.linalg.svd(X, full_matrices=True)
    Y = U[:, 2:]
    # Prepare the subspaces by picking the additional orthogonal vectors from Y
    A = jnp.column_stack((a1, Y[:, :D-1]))
    B = jnp.column_stack((b1, Y[:, D-1:2*D-2]))
    return A, B

two_subspaces_at_angle_jit = jit(two_subspaces_at_angle, static_argnums=(1,2,3))

def three_subspaces_at_angle(key, N, D, theta):
    """Returns ONBs for three subspaces at angle theta with each other

    Args:
        key: a PRNG key used as the random key.
        N (int): Dimension of the ambient space
        D (int): Dimension of the low dimensional subspace
        theta (float): Smallest principal angle (in degrees) between the three subspaces

    Returns:
        (jax.numpy.ndarray, jax.numpy.ndarray, jax.numpy.ndarray): A tuple consiting of three
        ONBs for the three subspaces

    Example:
        >>> rkey = random.PRNGKey(1)
        >>> N = 20
        >>> D = 4
        >>> theta = 15
        >>> A, B, C = three_subspaces_at_angle(rkey, N, D, theta)
        >>> print(A.shape, B.shape, C.shape)
        (20, 4) (20, 4) (20, 4)
        >>> from cr.nimble.subspaces import smallest_principal_angles_deg
        >>> angles = smallest_principal_angles_deg(jnp.array([A, B, C]))
        >>> print(jnp.round(angles, 2))
        [[ 0. 15. 15.]
        [15.  0. 15.]
        [15. 15.  0.]]
    """
    # Convert theta to radians
    theta = jnp.deg2rad(theta)
    # Draw three random vectors in the ambient space
    X = random.normal(key, (N, 3))
    # Orthogonalize them
    U, s, VH = jnp.linalg.svd(X, full_matrices=False)
    # Pick these three vectors
    a1 = U[:, 0]
    a2 = U[:, 1]
    a3 = U[:, 2]
    #  expected value of inner product between the vectors
    p = jnp.cos(theta)
    # linear combination terms for the first vector of the second space
    c1 = p
    s1 = jnp.sqrt(1 - c1**2)
    # first vector for second space
    b1 = s1 * a2 + c1 * a1
    # first vector for third space
    c1_1 = p
    c1_2 = p * (1 - p) / jnp.sqrt(1 - p**2)
    c1_3 = jnp.sqrt(1 - c1_1**2 - c1_2**2)
    c1 = c1_1 * a1 + c1_2 * a2 + c1_3 * a3
    # put these vectors together
    X = jnp.column_stack((a1, b1, c1))
    # Find the orthogonal complement of X 
    U, s, VH = jnp.linalg.svd(X, full_matrices=True)
    Y = U[:, 3:]
    # Prepare the subspaces by picking the additional orthogonal vectors from Y
    A = jnp.column_stack((a1, Y[:, :D-1]))
    B = jnp.column_stack((b1, Y[:, D-1:2*D-2]))
    C = jnp.column_stack((c1, Y[:, 2*D-2:3*D-3]))
    return A, B, C

three_subspaces_at_angle_jit = jit(three_subspaces_at_angle, static_argnums=(1,2,3))


import pytest

import numpy as np
from numpy.testing import (assert_almost_equal, assert_allclose, assert_,
                           assert_equal, assert_raises, assert_raises_regex,
                           assert_array_equal, assert_warns)


from jax.config import config
config.update("jax_enable_x64", True)


import jax
import jax.numpy as jnp
from jax import random, lax, vmap, jit

import cr.nimble as cnb

rtol = 1e-8 if jax.config.jax_enable_x64 else 1e-6
atol = 1e-7 if jax.config.jax_enable_x64 else 1e-5
decimal_cmp = 7 if jax.config.jax_enable_x64 else 6

float_type = jnp.float64 if jax.config.jax_enable_x64 else jnp.float32
complex_type = jnp.complex128 if jax.config.jax_enable_x64 else jnp.complex64



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility functions for ND arrays
"""

from jax import jit
import jax.numpy as jnp

from cr.nimble import promote_arg_dtypes


def arr_largest_index(x):
    """Returns the unraveled index of the largest entry (by magnitude) in an n-d array

    Args:
        x (jax.numpy.ndarray): An nd-array

    Returns:
        tuple: n-dim index of the largest entry in x 
    """
    x = jnp.asarray(x)
    return jnp.unravel_index(jnp.argmax(jnp.abs(x)), x.shape)

def arr_l1norm(x):
    """Returns the l1-norm of an array by flattening it

    Args:
        x (jax.numpy.ndarray): An nd-array

    Returns:
        (float): l1 norm of x 
    """
    x = jnp.asarray(x)
    x = promote_arg_dtypes(x)
    return jnp.sum(jnp.abs(x))


def arr_l2norm(x):
    """Returns the l2-norm of an array by flattening it
    """
    x = jnp.asarray(x)
    x = promote_arg_dtypes(x)
    return jnp.sqrt(jnp.abs(jnp.vdot(x, x)))

def arr_l2norm_sqr(x):
    """Returns the squared l2-norm of an array by flattening it
    """
    x = jnp.asarray(x)
    x = promote_arg_dtypes(x)
    return jnp.vdot(x, x)

def arr_vdot(x, y):
    """Returns the inner product of two arrays  by flattening it 
    """
    x = jnp.asarray(x)
    y = jnp.asarray(y)
    x, y = promote_arg_dtypes(x, y)
    return jnp.vdot(x, y)

@jit
def arr_rdot(x, y):
    """Returns the inner product Re(x^H, y) on two arrays by flattening them
    """
    x = jnp.asarray(x)
    y = jnp.asarray(y)
    x = jnp.ravel(x)
    y = jnp.ravel(y)
    if jnp.isrealobj(x) and jnp.isrealobj(y):
        # we can fall back to real inner product
        return jnp.sum(x * y)
    if jnp.isrealobj(x) or jnp.isrealobj(y):
        # 
        x = jnp.real(x)
        y = jnp.real(y)
        return jnp.sum(x * y)
    # both x and y are complex
    # compute x^H
    x = jnp.conjugate(x)    
    # compute x^H y
    prod = jnp.sum(x * y)
    # take the real part
    return jnp.real(prod)

@jit
def arr_rnorm_sqr(x):
    """Returns the squared norm of x using the real inner product Re(x^H, x)
    """
    return arr_rdot(x, x)

@jit
def arr_rnorm(x):
    """Returns the norm of x using the real inner product Re(x^H, x)
    """
    return jnp.sqrt(arr_rdot(x, x))

@jit
def arr2vec(x):
    """Converts an nd array to a vector
    """
    x = jnp.asarray(x)
    return jnp.ravel(x)


@jit
def log_pos(x):
    """Computes log with the assumption that x values are positive.
    """
    return jnp.log(jnp.maximum(x, jnp.finfo(float).eps))



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Some basic linear transformations
"""

import jax
import jax.numpy as jnp

from cr.nimble import promote_arg_dtypes
from cr.nimble import to_row_vec, to_col_vec


def householder_vec(x):
    """Computes a Householder vector for :math:`x`

    GVL4: Algorithm 5.1.1
    """
    x = promote_arg_dtypes(x)
    m = len(x)
    if m == 1:
        return jnp.array(0), jnp.array(0)
    x_1 = x[0]
    x_rest = x[1:]
    sigma = x_rest.T @ x_rest
    v = jnp.hstack((1, x_rest))

    def non_zero_sigma(v):
        mu = jnp.sqrt(x_1*x_1 + sigma)
        v_1 = jax.lax.cond(x_1 >= 0, 
            lambda _: x_1 - mu, 
            lambda _: -sigma/(x_1 + mu), 
            operand=None)
        v = v.at[0].set(v_1)
        beta = 2. * v_1 * v_1 / (sigma + v_1 * v_1)
        v = v / v_1
        return v, beta

    def zero_sigma(v):
        beta = jax.lax.cond(x_1 >= 0, lambda _: 0., lambda _: -2., operand=None)
        return v, beta

    v, beta = jax.lax.cond(sigma == 0, zero_sigma, non_zero_sigma, operand=v)
    return v , beta


def householder_matrix(x):
    """Computes a Householder refection matrix for :math:`x`
    """
    v, beta = householder_vec(x)
    return jnp.eye(len(x)) - beta * jnp.outer(v, v)


def householder_premultiply(v, beta, A):
    """Pre-multiplies a Householder reflection defined by :math:`v, beta` to a matrix A, PA
    """
    assert v.ndim == 1
    assert A.ndim == 2
    vt = to_row_vec(v)
    v = to_col_vec(v)
    return A - (beta * v) @(vt @ A)

def householder_postmultiply(v, beta, A):
    """Post-multiplies a Householder reflection defined by :math:`v, beta` to a matrix A, AP
    """
    assert v.ndim == 1
    assert A.ndim == 2
    vt = to_row_vec(v)
    v = to_col_vec(v)
    return A - (A @ v) @ (beta * vt)


def householder_vec_(x):
    """Computes a Householder vector for :math:`x`
    """
    x = promote_arg_dtypes(x)
    m = len(x)
    if m == 1:
        return jnp.array(0), jnp.array(0)
    x_1 = x[0]
    x_rest = x[1:]
    sigma = x_rest.T @ x_rest
    v = jnp.hstack((1, x_rest))
    
    if sigma == 0:
        if x_1 >= 0:
            beta = 0
        else:
            beta = -2
    else:
        mu = jnp.sqrt(x_1*x_1 + sigma)
        if x_1 <= 0:
            v = v.at[0].set(x_1 - mu)
        else:
            v = v.at[0].set(-sigma/(x_1 + mu))
        v_1 = v[0]
        beta = 2 * v_1 * v_1 / (sigma + v_1 * v_1)
        v = v / v_1
    return v , beta


def householder_ffm_jth_v_beta(A,j):
    """GVL4 EQ 5.1.4 v, beta calculation
    """
    v = A[j+1:, j]
    ms = v.T @ v
    beta = 2/(1 + ms)
    v = jnp.hstack((1, v))
    return v, beta

def householder_ffm_premultiply(A, C):
    """
    Computes Q^T C where Q is stored in its factored form in A.

    Each column j, of A contains the essential part of the j-th
    Householder vector.

    GVL4 EQ 5.1.4
    """
    m, n = A.shape
    for j in range(n):
        v, beta = householder_ffm_jth_v_beta(A, j)
        C2 = householder_premultiply(v, beta, C[j:,:])
        C = C.at[j:, :].set(C2)
    return C

def householder_ffm_backward_accum(A, k):
    """
    Computes k columns of Q from the factored form representation of Q stored in A.

    GVL4 EQ 5.1.5
    """
    m, n = A.shape
    Q = jnp.eye(m,k)
    for j in range(n-1, -1, -1):
        v, beta = householder_ffm_jth_v_beta(A, j)
        QQ = householder_premultiply(v, beta, Q[j:,j:])
        Q = Q.at[j:, j:].set(QQ)
    return Q


def householder_ffm_to_wy(A):
    """
    Computes the WY representation of Q such that Q = I_m - W Y^T from the factored form representation

    GVL4 algorithm 5.1.2
    """
    m, r = A.shape
    v, beta = householder_ffm_jth_v_beta(A, 0)
    v = to_col_vec(v)
    Y = v 
    W = beta * v
    for j in range (1, r-1):
        v, beta = householder_ffm_jth_v_beta(A, j)
        v = to_col_vec(v)
        v2 = jnp.vstack((jnp.zeros(j), v))
        z = beta * (v2  - (W @ Y[j:,:].T)@v)
        W = jnp.hstack((W, z))
        Y = jnp.hstack((Y, v2))
    return W, Y

def householder_qr_packed(A):
    """Computes the QR = A factorization of A using Householder reflections. Returns packed factorization.

    Algorithm 5.2.1
    """
    A = promote_arg_dtypes(A)
    m, n = A.shape
    assert m >= n
    for j in range(n-1):
        x = A[j:, j]
        v, beta = householder_vec(x)
        A2  = householder_premultiply(v, beta, A[j:, j:])
        A = A.at[j:, j:].set(A2)
        # place the essential part of the Householder vector
        A = A.at[j+1:,j].set(v[1:])
    return A

def householder_split_qf_r(A):
    """Splits a packed QR factorization into QF and R
    """
    # The upper triangular part is R
    R = jnp.triu(A)
    # The remaining lower triangular part of A is the factored form representation of Q
    QF = jnp.tril(A[:,:-1], -1)
    return QF, R

def householder_qr(A):
    """Computes the QR = A factorization of A using Householder reflections

    Algorithm 5.2.1
    """
    m, n = A.shape
    A = householder_qr_packed(A)
    QF , R = householder_split_qf_r(A)
    Q = householder_ffm_backward_accum(QF, n)
    return Q, R





# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import NamedTuple
import math

from jax import lax, jit, vmap, random
import jax.numpy as jnp
from jax.numpy.linalg import norm


import cr.nimble as cnb

from .lanbpro import (LanBDOptions, LanBProState, 
    lanbpro_options_init, lanbpro_init, lanbpro_iteration, lanbpro)

from .bdsqr import bdsqr
from .lansvd_utils import refine_bounds

class LanSVDState(NamedTuple):
    """State for Lan SVD iterations
    """
    bpro_state: LanBProState
    "State for the LanBPro algorithm"
    n_converged : int = 0
    "Number of converged eigen values"

def lansvd_simple(A, k, p0):
    """Returns the k largest singular values and corresponding vectors
    """
    m, n = A.shape
    # maximum number of U/V columns
    lanmax = min(m, n)
    lanmax = min (lanmax, max(20, k*4))
    assert k <= lanmax, "k must be less than lanmax"

    eps = jnp.finfo(float).eps
    tol = 16*eps
    n_converged = 0
    # number of iterations for LanBPro algorithm
    j = min(k + max(8, k), lanmax)
    options = lanbpro_options_init(lanmax)
    state  = lanbpro_init(A, lanmax, p0, options)
    # carry out the lanbpro iterations
    state = lax.fori_loop(1, j, 
        lambda i, state: lanbpro_iteration(A, state, options),
        state)
    # norm of the residual 
    res_norm = norm(state.p)
    # compute SVD of the bidiagonal matrix ritz values and vectors
    P, S, Qh, bot = bdsqr(state.alpha, state.beta, j)
    # estimate of A norm 
    anorm = S[0]
    # simple error bounds on singular values
    bnd = res_norm * jnp.abs(bot)
    # now refine the bounds
    bnd = refine_bounds(S**2, bnd, n*eps*anorm)
    # count the number of converged singular values
    converged = jnp.less(bnd, jnp.abs(S))
    # make sure that all indices beyond min(j,k) are marked as non-converged
    converged = converged.at[min(j,k):].set(False)
    # find the index of first non-converged singular value
    n_converged = jnp.argmin(converged)
    n_converged = jnp.where(converged[n_converged], len(converged), n_converged)
    U = state.U[:, :j]
    V = state.V[:, :j]
    # keep only the first k ritz vectors
    P = P[:, :k]
    Q = Qh.T[:, :k]
    U = state.U[:, :j] @ P[:j, :]
    V = state.V[:, :j] @ Q
    return U, S[:k], V, bnd, n_converged, state



lansvd_simple_jit = jit(lansvd_simple, static_argnums=(1,))



# Copyright 2021 CR.Sparse Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
from jax import random, jit

from cr.nimble import promote_arg_dtypes
from cr.nimble import arr_l2norm_sqr

def awgn_at_snr(key, signal, snr):
    """Generates noise for the signal at the specified SNR
    """
    signal = jnp.asarray(signal)
    signal = promote_arg_dtypes(signal)
    n = signal.size
    energy = arr_l2norm_sqr(signal)
    mean_energy = energy / n
    mean_energy_db = 10*jnp.log10(mean_energy)
    noise_mean_energy_db = mean_energy_db - snr
    sigma = 10**(noise_mean_energy_db/20)
    noise = sigma * random.normal(key, signal.shape)
    return noise


# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from jax.scipy.linalg import solve_triangular

def solve_Lx_b(L, b):
    """
    Solves the system L x = b using back substitution
    """
    return solve_triangular(L, b, lower=True)

def solve_LTx_b(L, b):
    """
    Solves the system L^T x = b using back substitution
    """
    return solve_triangular(L, b, lower=True, trans='T')

def solve_Ux_b(U, b):
    """
    Solves the system U x = b using back substitution
    """
    return solve_triangular(U, b)

def solve_UTx_b(U, b):
    """
    Solves the system U^T x = b using back substitution
    """
    return solve_triangular(U, b, trans='T')


def solve_spd_chol(L, b):
    """
    Solves a symmetric positive definite system A x = b
    where A = L L'
    """
    # We have to solve L L' x = b
    # We first solve L u = b
    u = solve_Lx_b(L, b)
    # We now solve L' x = u
    x = solve_LTx_b(L, u)
    return x



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Some basic linear transformations
"""

import jax
import jax.numpy as jnp

from cr.nimble import promote_arg_dtypes

def point2d(x,y):
    """A point in 2D vector space"""
    return jnp.array([x+0.,y])

def vec2d(x,y):
    """A vector in 2D vector space"""
    return jnp.array([x+0.,y])

def rotate2d_cw(theta):
    """Construct an operator that rotates a 2D vector by angle :math:`\theta` clock-wise
    """
    Q  = jnp.array([[jnp.cos(theta), jnp.sin(theta)],
        [-jnp.sin(theta), jnp.cos(theta)]])
    return Q

def rotate2d_ccw(theta):
    """Construct an operator that rotates a 2D vector by angle :math:`\theta` counter-clock-wise
    """
    Q  = jnp.array([[jnp.cos(theta), -jnp.sin(theta)],
        [jnp.sin(theta), jnp.cos(theta)]])
    return Q

def reflect2d(theta):
    """Construct an operator that reflects a 2D vector across a line defined at angle :math:`\theta/2`
    """
    R  = jnp.array([[jnp.cos(theta), jnp.sin(theta)],
        [jnp.sin(theta), -jnp.cos(theta)]])
    return R



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp
from jax import jit, lax

from jax.numpy.linalg import norm

from .svd_utils import singular_values
from .array import hermitian
from .matrix import AH_v

def orth_complement(A, B):
    """Returns the orthogonal complement of A in B
    """
    rank_a = A.shape[1]
    C = jnp.hstack([A, B])
    Q, R = jnp.linalg.qr(C)
    return Q[:, rank_a:]

def principal_angles_cos(A, B):
    """Returns the cosines of principal angles between two subspaces 

    Args:
        A (jax.numpy.ndarray): ONB for the first subspace
        B (jax.numpy.ndarray): ONB for the second subspace

    Returns:
        (jax.numpy.ndarray): The list of principal angles between two subspaces 
            from smallest to the largest.
    """
    AH = jnp.conjugate(A.T)
    M = AH @ B
    s = singular_values(M)
    # ensure that the singular values are below 1
    return jnp.minimum(1, s)

principal_angles_cos_jit = jit(principal_angles_cos)

def principal_angles_rad(A, B):
    """Returns the principal angles between two subspaces in radians

    Args:
        A (jax.numpy.ndarray): ONB for the first subspace
        B (jax.numpy.ndarray): ONB for the second subspace

    Returns:
        (jax.numpy.ndarray): The list of principal angles between two subspaces 
            from smallest to the largest.
    """
    angles = principal_angles_cos(A, B)
    return jnp.arccos(angles)

principal_angles_rad_jit = jit(principal_angles_rad)

def principal_angles_deg(A, B):
    """Returns the principal angles between two subspaces in degrees

    Args:
        A (jax.numpy.ndarray): ONB for the first subspace
        B (jax.numpy.ndarray): ONB for the second subspace

    Returns:
        (jax.numpy.ndarray): The list of principal angles between two subspaces 
            from smallest to the largest.
    """
    angles = principal_angles_rad(A, B)
    return jnp.rad2deg(angles)

principal_angles_deg_jit = jit(principal_angles_deg)

def smallest_principal_angle_cos(A, B):
    """Returns the cosine of smallest principal angle between two subspaces 

    Args:
        A (jax.numpy.ndarray): ONB for the first subspace
        B (jax.numpy.ndarray): ONB for the second subspace

    Returns:
        (float): Cosine of the smallest principal angle between the two subspaces
    """
    angles = principal_angles_cos(A, B)
    return angles[0]

smallest_principal_angle_cos_jit = jit(smallest_principal_angle_cos)

def smallest_principal_angle_rad(A, B):
    """Returns the smallest principal angle between two subspaces in radians

    Args:
        A (jax.numpy.ndarray): ONB for the first subspace
        B (jax.numpy.ndarray): ONB for the second subspace

    Returns:
        (float): The smallest principal angle between the two subspaces in radians
    """
    angle = smallest_principal_angle_cos(A, B)
    return jnp.arccos(angle)

smallest_principal_angle_rad_jit = jit(smallest_principal_angle_rad)

def smallest_principal_angle_deg(A, B):
    """Returns the smallest principal angle between two subspaces in degrees

    Args:
        A (jax.numpy.ndarray): ONB for the first subspace
        B (jax.numpy.ndarray): ONB for the second subspace

    Returns:
        (float): The smallest principal angle between the two subspaces in degrees
    """
    angle = smallest_principal_angle_rad(A, B)
    return jnp.rad2deg(angle)

smallest_principal_angle_deg_jit = jit(smallest_principal_angle_deg)

def smallest_principal_angles_cos(subspaces):
    """Returns the smallest principal angles between each pair of subspaces

    Args:
        A (jax.numpy.ndarray): An array of ONBs for the subspaces

    Returns:
        (jax.numpy.ndarray): A symmetric matrix containing the cosine of the 
            smallest principal angles between each pair of subspaces

    Further reading on implementation:

    * `Vectorizing computations on pairs of elements in an nd-array <https://towardsdatascience.com/vectorizing-computations-on-pairs-of-elements-in-an-nd-array-326b5a648ad6>`_
    * `SO: How to vectorize a 2 level loop in NumPy <https://stackoverflow.com/questions/69391894/how-to-vectorize-a-2-level-loop-in-numpy>`_
    """
    subspaces = jnp.asarray(subspaces)
    # Number of subspaces
    k = subspaces.shape[0]
    # Indices for upper triangular matrix
    i, j = jnp.triu_indices(k, k=1)
    # prepare all the possible pairs of A and B
    A = subspaces[i]
    B = subspaces[j]
    AH = jnp.conjugate(jnp.transpose(A, axes=(0,2,1)))
    M = jnp.matmul(AH, B)
    s = jnp.linalg.svd(M, compute_uv=False)
    # keep only the first index
    s = s[:, 0]
    # prepare the returning matrix
    r = jnp.eye(k)
    r = r.at[i, j].set(s)
    r = r + r.T - jnp.eye(k)
    # make sure that there is no overflow
    r = jnp.minimum(r, 1.)
    return r


smallest_principal_angles_cos_jit = jit(smallest_principal_angles_cos)


def smallest_principal_angles_rad(subspaces):
    """Returns the smallest principal angles between each pair of subspaces in radians

    Args:
        A (jax.numpy.ndarray): An array of ONBs for the subspaces

    Returns:
        (jax.numpy.ndarray): A symmetric matrix containing the 
            smallest principal angles between each pair of subspaces in radians
    """
    result = smallest_principal_angles_cos(subspaces)
    return jnp.arccos(result)

smallest_principal_angles_rad_jit = jit(smallest_principal_angles_rad)

def smallest_principal_angles_deg(subspaces):
    """Returns the smallest principal angles between each pair of subspaces in degrees

    Args:
        A (jax.numpy.ndarray): An array of ONBs for the subspaces

    Returns:
        (jax.numpy.ndarray): A symmetric matrix containing the 
            smallest principal angles between each pair of subspaces in degrees
    """
    result = smallest_principal_angles_rad(subspaces)
    return jnp.rad2deg(result)

smallest_principal_angles_deg_jit = jit(smallest_principal_angles_deg)


def subspace_distance(A, B):
    r"""Returns the Grassmannian distance between two subspaces

    Args:
        A (jax.numpy.ndarray): ONB for the first subspace
        B (jax.numpy.ndarray): ONB for the second subspace

    Returns:
        (float): Distance between the two subspaces

    the `Grassmannian <https://en.wikipedia.org/wiki/Grassmannian>`_ 
    is a space that parameterizes  all  k dimensional linear 
    subspaces of a vector space V. 
    A `metric <https://math.stackexchange.com/questions/198111/distance-between-real-finite-dimensional-linear-subspaces>`_
    can be defined over this space. We can use this metric
    to compute the distance between two subspaces.
    """
    # Compute the projection operators for the two subspaces
    PA = A @ jnp.conjugate(A.T)
    PB = B @ jnp.conjugate(B.T)
    # Difference between the projection operators
    D = PA  - PB
    # Return the operator norm of D
    return jnp.linalg.norm(D)

subspace_distance_jit = jit(subspace_distance)


def project_to_subspace(U, v):
    """Projects a vector to a subspace

    Args:
        U (jax.numpy.ndarray): ONB for the subspace
        v (jax.numpy.ndarray): A vector in the ambient space

    Returns:
        (jax.numpy.ndarray): Projection of v onto the subspace spanned by U

    Example:
        >>> A = jnp.eye(6)[:, :3]
        >>> v = jnp.arange(6) + 0.
        >>> u = project_to_subspace(A, v)
        >>> print(A)
        [[1. 0. 0.]
        [0. 1. 0.]
        [0. 0. 1.]
        [0. 0. 0.]
        [0. 0. 0.]
        [0. 0. 0.]]
        >>> print(v)
        [0. 1. 2. 3. 4. 5.]
        >>> print(u)
        [0. 1. 2. 0. 0. 0.]
    """
    UHv = AH_v(U, v)
    return U  @ UHv

def is_in_subspace(U, v):
    """Checks whether a vector v is in the subspace spanned by an ONB U or not

    Args:
        U (jax.numpy.ndarray): ONB for the subspace
        v (jax.numpy.ndarray): A vector in the ambient space

    Returns:
        (bool): True if v lies in the subspace spanned by U, False otherwise

    Example:
        >>> A = jnp.eye(6)[:, :3]
        >>> v = jnp.arange(6) + 0.
        >>> print(is_in_subspace(A, v))
        False
        >>> u = project_to_subspace(A, v)
        >>> print(is_in_subspace(A, u))
        True
    """
    # Compute the projection
    p = project_to_subspace(U, v)
    # Compute the error
    e = p - v
    nv = norm(v)
    ne = norm(e)
    return ne <= 1e-6 * nv


# Copyright 2021 CR.Sparse Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
import jax.numpy as jnp

from cr.nimble import sqr_norms_l2_cw, sqr_norms_l2_rw, norms_l2_cw, sqr_norm_l2

norm = jnp.linalg.norm


class SignalsComparison:

    def __init__(self, references, estimates):
        if references.ndim == 1:
            references = jnp.expand_dims(references, 1)
        if estimates.ndim == 1:
            estimates = jnp.expand_dims(estimates, 1)
        self.references = references
        self.estimates  = estimates
        self.differences = references - estimates

    @property
    def reference_norms(self):
        return norms_l2_cw(self.references)

    @property
    def estimate_norms(self):
        return norms_l2_cw(self.estimates)

    @property
    def difference_norms(self):
        return norms_l2_cw(self.differences)

    @property
    def reference_energies(self):
        return sqr_norms_l2_cw(self.references)

    @property
    def estimate_energies(self):
        return sqr_norms_l2_cw(self.estimates)

    @property
    def difference_energies(self):
        return sqr_norms_l2_cw(self.differences)

    @property
    def error_to_signal_norms(self):
        a_norms = self.reference_norms
        diff_norms = self.difference_norms
        return diff_norms / a_norms
    
    @property
    def signal_to_noise_ratios(self):
        a_energies = self.reference_energies
        diff_energies = self.difference_energies
        ratios = a_energies / diff_energies
        return 10*jnp.log10(ratios)

    @property
    def cum_reference_norm(self):
        return norm(self.references, 'fro')

    @property
    def cum_estimate_norm(self):
        return norm(self.estimates, 'fro')

    @property
    def cum_difference_norm(self):
        return norm(self.differences, 'fro')

    @property
    def cum_error_to_signal_norm(self):
        a_norm = self.cum_reference_norm
        err_norm = self.cum_difference_norm
        ratio = err_norm / a_norm
        return ratio

    @property
    def cum_signal_to_noise_ratio(self):
        a_norm = self.cum_reference_norm
        err_norm = self.cum_difference_norm
        ratio = a_norm  / err_norm
        return 20 * jnp.log10(ratio)

    def summarize(self):
        # if self.references.ndim == 2:
        #     
        # else:
        #     n, s = self.references.shape, 1
        n, s = self.references.shape
        print(f'Dimensions: {n}')
        print(f'Signals: {s}')
        print(f'Combined reference norm: {self.cum_reference_norm:.3f}')
        print(f'Combined estimate norm: {self.cum_estimate_norm:.3f}')
        print(f'Combined difference norm: {self.cum_difference_norm:.3f}')
        print(f'Combined SNR: {self.cum_signal_to_noise_ratio:.2f} dB')



def snrs_cw(A, B):
    ref_energies = sqr_norms_l2_cw(A)
    diff_energies = sqr_norms_l2_cw(A-B)
    ratios = ref_energies / diff_energies
    return 10*jnp.log10(ratios)

def snrs_rw(A, B):
    ref_energies = sqr_norms_l2_rw(A)
    diff_energies = sqr_norms_l2_rw(A-B)
    ratios = ref_energies / diff_energies
    return 10*jnp.log10(ratios)

def snr(signal, noise):
    s = sqr_norm_l2(signal)
    n = sqr_norm_l2(noise)
    ratio = s/n
    return 10*jnp.log10(ratio)




import math
from functools import partial

import numpy as np
import scipy

import jax
import jax.numpy as jnp
from jax import random
from jax import jit

from .norm import normalize_l2_cw
from .util import promote_arg_dtypes
from .array import hermitian

def gaussian_mtx(key, N, D, normalize_atoms=True):
    """A dictionary/sensing matrix where entries are drawn independently from normal distribution.

    Args:
        key: a PRNG key used as the random key.
        N (int): Number of rows of the sensing matrix 
        D (int): Number of columns of the sensing matrix
        normalize_atoms (bool): Whether the columns of sensing matrix are normalized 
          (default True)

    Returns:
        (jax.numpy.ndarray): A Gaussian sensing matrix of shape (N, D)

    Example:

        >>> from jax import random
        >>> import cr.nimble as cnb
        >>> m, n = 8, 16
        >>> Phi = cnb.gaussian_mtx(random.PRNGKey(0), m, n)
        >>> print(Phi.shape)
        (8, 16)
        >>> print(cnb.norms_l2_cw(Phi))
        [1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1.]
    """
    shape = (N, D)
    dict = random.normal(key, shape)
    if normalize_atoms:
        dict = normalize_l2_cw(dict)
    else:
        sigma = math.sqrt(N)
        dict = dict / sigma
    return dict


def _pascal_lower(n):
    A = jnp.empty((n, n), dtype=jnp.int32)
    A = A.at[0, :].set(0)
    A = A.at[:, 0].set(1)
    for i in range(1, n):
        for j in range(1, i+1):
            A = A.at[i, j].set(A[i-1, j] + A[i-1, j-1])
    return A

def _pascal_sym(n):
    A = jnp.empty((n, n), dtype=jnp.int32)
    A = A.at[0, :].set(1)
    A = A.at[:, 0].set(1)
    for i in range(1, n):
        for j in range(1, n):
            A = A.at[i, j].set(A[i-1, j] + A[i, j-1])
    return A

def pascal(n, symmetric=False):
    """Returns a pascal matrix of size n \times n
    """
    if symmetric:
        return _pascal_sym(n)
    else:
        return _pascal_lower(n)

pascal_jit = jax.jit(pascal, static_argnums=(0, 1))



# Copyright 2021 CR.Sparse Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Similarity measures
"""

from jax import jit
import jax.numpy as jnp

def dist_to_gaussian_sim(dist, sigma):
    """Computes the Gaussian similarities for given distances
    """
    d = dist**2 / (2 * sigma**2)
    return jnp.exp(-d)

def sqr_dist_to_gaussian_sim(sqr_dist, sigma):
    """Computes the Gaussian similarities for given squared distances
    """
    d = sqr_dist / (2 * sigma**2)
    return jnp.exp(-d)

def eps_neighborhood_sim(dist, threshold):
    """Computes the epsilon neighborhood similarity
    """
    return dist < threshold


# Copyright 2022-Present CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def norm_freq(frequency, sampling_rate):
    """Returns the normalized frequency

    The Nyquist frequency is the half of the sampling rate.
    In normalized range, the Nyquist frequency has a value of 1.
    If sampling rate is 200 Hz and signal frequency is
    20 Hz, then Nyquist frequency is 100 Hz and the
    normalized frequency is 0.2.

    Args:
        frequency (float): Frequency in Hz.
        sampling_rate (float): Sampling rate of signal in Hz.

    Returns:
        float: Normalized sampling frequency
    """
    return 2.0 * frequency / sampling_rate



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cr.nimble._src.subspaces import (
    orth_complement,
    project_to_subspace,
    is_in_subspace,
    principal_angles_cos,
    principal_angles_cos_jit,
    principal_angles_rad,
    principal_angles_rad_jit,
    principal_angles_deg,
    principal_angles_deg_jit,
    smallest_principal_angle_cos,
    smallest_principal_angle_cos_jit,
    smallest_principal_angle_rad,
    smallest_principal_angle_rad_jit,
    smallest_principal_angle_deg,
    smallest_principal_angle_deg_jit,
    smallest_principal_angles_cos,
    smallest_principal_angles_cos_jit,
    smallest_principal_angles_rad,
    smallest_principal_angles_rad_jit,
    smallest_principal_angles_deg,
    smallest_principal_angles_deg_jit,
    subspace_distance,
)